import pint

reg = pint.UnitRegistry()

# Xi bytes are not known to pint
reg.define("Ei = 1 EiB")
reg.define("Pi = 1 PiB")
reg.define("Ti = 1 TiB")
reg.define("Gi = 1 GiB")
reg.define("Mi = 1 MiB")
reg.define("Ki = 1 KiB")
reg.define("KB = 1000 B")

# globally define or register units

bytes_u: pint.Quantity = reg.byte


def parse(s: str) -> pint.Quantity:
    return reg.parse_expression(s)



from datetime import timedelta
from typing import Optional, ClassVar, Union, Literal, Any

from attrs import define

from resotolib.json import to_json, from_json, is_primitive_or_primitive_union, sort_json
from resotolib.utils import utc


@define
class Foo:
    static: ClassVar[str] = "static"
    a: str
    b: int
    c: Optional[str] = None
    d: Union[str, int] = "test"
    inner: Optional["Foo"] = None
    _private: str = "private"

    @property
    def foo(self) -> str:
        return "foo"


def test_complex() -> None:
    foo = Foo("foo", 42, "bar")
    assert to_json(foo) == {"a": "foo", "b": 42, "c": "bar", "d": "test", "inner": None}
    json = to_json(foo)
    assert json == {"a": "foo", "b": 42, "c": "bar", "d": "test", "inner": None}
    again = from_json(json, Foo)
    assert isinstance(again, Foo)
    assert again == foo
    # strip attrs based on string or list of strings
    assert to_json(foo, strip_attr="a") == {"b": 42, "c": "bar", "d": "test", "inner": None}
    assert to_json(foo, strip_attr=["a", "b", "c"]) == {"d": "test", "inner": None}
    # do not strip nulls
    assert to_json(foo, strip_nulls=True) == {"a": "foo", "b": 42, "c": "bar", "d": "test"}


def test_predefined() -> None:
    def roundtrip(value: Any) -> Any:
        again = from_json(to_json(value), type(value))
        assert to_json(again) == to_json(value)

    roundtrip(utc())
    roundtrip(utc().date())
    roundtrip(timedelta(seconds=42))
    roundtrip(Foo("foo", 42, "bar"))


def test_primitive_union() -> None:
    # simple types
    assert is_primitive_or_primitive_union(str) is True
    assert is_primitive_or_primitive_union(int) is True
    assert is_primitive_or_primitive_union(bool) is True
    assert is_primitive_or_primitive_union(int) is True
    assert is_primitive_or_primitive_union(float) is True
    assert is_primitive_or_primitive_union(type(None)) is True
    assert is_primitive_or_primitive_union(Foo) is False
    # literal types
    assert is_primitive_or_primitive_union(Literal["test"]) is True
    # union types
    assert is_primitive_or_primitive_union(Union[str, int, None]) is True
    assert is_primitive_or_primitive_union(Union[str, int, Foo]) is False
    assert is_primitive_or_primitive_union(Union[float]) is True
    assert is_primitive_or_primitive_union(Optional[str]) is True
    assert is_primitive_or_primitive_union(Optional[int]) is True
    assert is_primitive_or_primitive_union(Optional[Foo]) is False


def test_json_sort() -> None:
    simple = {"c": [3, 2, 1], "b": 2, "a": 1}
    simple2 = {"c": [3, 2, 4], "b": 3, "a": 2}
    simple3 = {"c": [5, 2, 3], "b": 4, "a": 3}
    assert sort_json(simple) == {"a": 1, "b": 2, "c": [3, 2, 1]}
    assert sort_json(simple, sort_list=True) == {"a": 1, "b": 2, "c": [1, 2, 3]}
    complex = {"d": [simple2, simple3, simple], "a": {"c": simple3, "b": simple2, "a": simple}}
    assert sort_json(complex, sort_list=True) == {
        "a": {
            "a": {"a": 1, "b": 2, "c": [1, 2, 3]},
            "b": {"a": 2, "b": 3, "c": [2, 3, 4]},
            "c": {"a": 3, "b": 4, "c": [2, 3, 5]},
        },
        "d": [{"a": 2, "b": 3, "c": [2, 3, 4]}, {"a": 3, "b": 4, "c": [2, 3, 5]}, {"a": 1, "b": 2, "c": [1, 2, 3]}],
    }



import os
import base64
import certifi
from ipaddress import (
    ip_address,
    ip_network,
    IPv4Address,
    IPv6Address,
    IPv4Network,
    IPv6Network,
)

from cryptography.hazmat._oid import ExtendedKeyUsageOID

from resotolib.utils import get_local_hostnames, get_local_ip_addresses
from datetime import datetime, timedelta, timezone
from cryptography import x509
from cryptography.x509.oid import NameOID
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.asymmetric.rsa import (
    RSAPrivateKey,
    RSAPublicKey,
    generate_private_key,
)
from cryptography.x509.base import Certificate, CertificateSigningRequest
from cryptography.exceptions import InvalidSignature
from typing import List, Optional, Tuple, Union, Dict, Any


def gen_rsa_key(key_size: int = 2048) -> RSAPrivateKey:
    return generate_private_key(public_exponent=65537, key_size=key_size, backend=default_backend())


def bootstrap_ca(
    days_valid: int = 3650,
    common_name: str = "Resoto Root CA",
    organization_name: str = "Some Engineering Inc.",
    path_length: int = 2,
) -> Tuple[RSAPrivateKey, Certificate]:
    ca_key = gen_rsa_key()
    subject = issuer = x509.Name(
        [
            x509.NameAttribute(NameOID.ORGANIZATION_NAME, organization_name),
            x509.NameAttribute(NameOID.COMMON_NAME, common_name),
        ]
    )
    ca_cert = (
        x509.CertificateBuilder()
        .subject_name(subject)
        .issuer_name(issuer)
        .public_key(ca_key.public_key())
        .serial_number(x509.random_serial_number())
        .not_valid_before(datetime.now(tz=timezone.utc))
        .not_valid_after(datetime.now(tz=timezone.utc) + timedelta(days=days_valid))
        .add_extension(x509.BasicConstraints(ca=True, path_length=path_length), critical=True)
        .add_extension(
            x509.KeyUsage(
                digital_signature=False,
                key_encipherment=False,
                key_cert_sign=True,  # CA Cert is only allowed to sign other certs
                key_agreement=False,
                content_commitment=False,
                data_encipherment=False,
                crl_sign=True,  # and cert revocation lists
                encipher_only=False,
                decipher_only=False,
            ),
            critical=True,  # KeyUsage extension is critical to support
        )
        .sign(ca_key, hashes.SHA256(), default_backend())
    )
    return ca_key, ca_cert


def gen_csr(
    csr_key: RSAPrivateKey,
    *,
    common_name: str = "some.engineering",
    san_dns_names: Optional[List[str]] = None,
    san_ip_addresses: Optional[List[str]] = None,
    include_loopback: bool = True,
    connect_to_ips: Optional[List[str]] = None,
    discover_local_dns_names: bool = True,
    discover_local_ip_addresses: bool = True,
) -> CertificateSigningRequest:
    if san_dns_names is None:
        san_dns_names = []
    elif isinstance(san_dns_names, str):
        san_dns_names = [san_dns_names]
    if san_ip_addresses is None:
        san_ip_addresses = []
    elif isinstance(san_ip_addresses, str):
        san_ip_addresses = [san_ip_addresses]

    if discover_local_dns_names:
        san_dns_names = get_local_hostnames(
            include_loopback=include_loopback,
            san_ip_addresses=san_ip_addresses,
            san_dns_names=san_dns_names,
            connect_to_ips=connect_to_ips,
        )

    if discover_local_ip_addresses:
        san_ip_addresses = get_local_ip_addresses(
            include_loopback=include_loopback,
            san_ip_addresses=san_ip_addresses,
            connect_to_ips=connect_to_ips,
        )

    csr_build = x509.CertificateSigningRequestBuilder().subject_name(
        x509.Name([x509.NameAttribute(NameOID.COMMON_NAME, common_name)])
    )
    if len(san_dns_names) + len(san_ip_addresses) > 0:
        csr_build = csr_build.add_extension(
            x509.SubjectAlternativeName(
                [x509.DNSName(n) for n in san_dns_names] + [x509.IPAddress(make_ip(i)) for i in san_ip_addresses]
            ),
            critical=False,  # Optional extensions are not critical if unsupported
        )
    return csr_build.sign(csr_key, hashes.SHA256(), default_backend())


def sign_csr(
    csr: CertificateSigningRequest,
    ca_key: RSAPrivateKey,
    ca_cert: Certificate,
    days_valid: int = 365,
    server_auth: bool = True,
    client_auth: bool = True,
    key_usages: Optional[Dict[str, bool]] = None,
) -> Certificate:
    usage = key_usages or {}
    crt_build = (
        x509.CertificateBuilder()
        .subject_name(csr.subject)
        .issuer_name(ca_cert.subject)
        .public_key(csr.public_key())
        .serial_number(x509.random_serial_number())
        .not_valid_before(datetime.now(tz=timezone.utc))
        .not_valid_after(datetime.now(tz=timezone.utc) + timedelta(days=days_valid))
        .add_extension(
            x509.KeyUsage(
                digital_signature=usage.get("digital_signature", True),  # Server/client certs are allowed for
                key_encipherment=usage.get("key_encipherment", True),  # signatures and encrypting traffic.
                key_cert_sign=usage.get("key_cert_sign", False),
                key_agreement=usage.get("key_agreement", False),
                content_commitment=usage.get("content_commitment", False),
                data_encipherment=usage.get("data_encipherment", False),
                crl_sign=usage.get("crl_sign", False),
                encipher_only=usage.get("encipher_only", False),
                decipher_only=usage.get("decipher_only", False),
            ),
            critical=True,
        )
    )
    if server_auth or client_auth:
        key_usage = []
        if server_auth:
            key_usage.append(ExtendedKeyUsageOID.SERVER_AUTH)
        if client_auth:
            key_usage.append(ExtendedKeyUsageOID.CLIENT_AUTH)
        crt_build = crt_build.add_extension(x509.ExtendedKeyUsage(key_usage), critical=False)
    for extension in csr.extensions:
        if not isinstance(extension.value, x509.SubjectAlternativeName):
            continue
        crt_build = crt_build.add_extension(extension.value, critical=extension.critical)
    return crt_build.sign(ca_key, hashes.SHA256(), default_backend())


def write_csr_to_file(csr: CertificateSigningRequest, csr_path: str, rename: bool = True) -> None:
    tmp_csr_path = f"{csr_path}.tmp" if rename else csr_path
    with open(tmp_csr_path, "wb") as f:
        f.write(csr_to_bytes(csr))
    if rename:
        os.rename(tmp_csr_path, csr_path)


def write_cert_to_file(cert: Certificate, cert_path: str, rename: bool = True) -> None:
    tmp_cert_path = f"{cert_path}.tmp" if rename else cert_path
    with open(tmp_cert_path, "wb") as f:
        f.write(cert_to_bytes(cert))
    if rename:
        os.rename(tmp_cert_path, cert_path)


def gen_ca_bundle_bytes(certs: Union[Certificate, List[Certificate]], include_certifi: bool = True) -> bytes:
    content = bytearray()
    if include_certifi:
        content.extend(certifi.contents().encode())

    if isinstance(certs, Certificate):
        certs = [certs]

    for cert in certs:
        content.extend("\n".encode())
        content.extend(f"# Issuer: {cert.issuer.rfc4514_string()}\n".encode())
        content.extend(f"# Subject: {cert.subject.rfc4514_string()}\n".encode())
        label: str = cert.issuer.get_attributes_for_oid(NameOID.COMMON_NAME)[0].value  # type: ignore
        content.extend(f"# Label: {label}\n".encode())
        content.extend(f"# Serial: {cert.serial_number}\n".encode())
        md5 = cert_fingerprint(cert, "MD5")
        sha1 = cert_fingerprint(cert, "SHA1")
        sha256 = cert_fingerprint(cert, "SHA256")
        content.extend(f"# MD5 Fingerprint: {md5}\n".encode())
        content.extend(f"# SHA1 Fingerprint: {sha1}\n".encode())
        content.extend(f"# SHA256 Fingerprint: {sha256}\n".encode())
        content.extend(cert_to_bytes(cert))
    return bytes(content)


def write_ca_bundle(
    certs: Union[Certificate, List[Certificate]], cert_path: str, include_certifi: bool = True, rename: bool = True
) -> None:
    tmp_cert_path = f"{cert_path}.tmp" if rename else cert_path
    with open(tmp_cert_path, "wb") as f:
        f.write(gen_ca_bundle_bytes(certs, include_certifi))
    if rename:
        os.rename(tmp_cert_path, cert_path)


def write_key_to_file(
    key: RSAPrivateKey,
    key_path: str,
    passphrase: Optional[str] = None,
    rename: bool = True,
) -> None:
    tmp_key_path = f"{key_path}.tmp" if rename else key_path
    with open(os.open(tmp_key_path, os.O_CREAT | os.O_WRONLY | os.O_TRUNC, 0o600), "wb") as f:
        f.write(key_to_bytes(key, passphrase))
    if rename:
        os.rename(tmp_key_path, key_path)


def key_to_bytes(
    key: RSAPrivateKey,
    passphrase: Optional[str] = None,
) -> bytes:
    kwargs: Dict[str, Any] = {"encryption_algorithm": serialization.NoEncryption()}
    if passphrase is not None:
        kwargs["encryption_algorithm"] = serialization.BestAvailableEncryption(passphrase.encode())
    return key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        **kwargs,
    )


def csr_to_bytes(csr: CertificateSigningRequest) -> bytes:
    return csr.public_bytes(serialization.Encoding.PEM)


def cert_to_bytes(cert: Certificate) -> bytes:
    return cert.public_bytes(serialization.Encoding.PEM)


def load_csr_from_file(csr_path: str) -> CertificateSigningRequest:
    with open(csr_path, "rb") as f:
        csr = f.read()
    return load_csr_from_bytes(csr)


def load_cert_from_file(cert_path: str) -> Certificate:
    with open(cert_path, "rb") as f:
        cert = f.read()
    return load_cert_from_bytes(cert)


def load_key_from_file(key_path: str, passphrase: Optional[str] = None) -> RSAPrivateKey:
    with open(key_path, "rb") as f:
        key = f.read()
    return load_key_from_bytes(key, passphrase)


def load_csr_from_bytes(csr: bytes) -> CertificateSigningRequest:
    return x509.load_pem_x509_csr(csr, default_backend())


def load_cert_from_bytes(cert: bytes) -> Certificate:
    return x509.load_pem_x509_certificate(cert, default_backend())


def load_key_from_bytes(
    key: bytes, passphrase: Optional[str] = None, skip_rsa_key_validation: bool = False
) -> RSAPrivateKey:
    backend = default_backend()
    if passphrase is not None:
        passphrase = passphrase.encode()  # type: ignore
    return backend.load_pem_private_key(key, passphrase, skip_rsa_key_validation)  # type: ignore


def make_ip(ip: str) -> Union[IPv4Address, IPv6Address, IPv4Network, IPv6Network]:
    if "/" in ip:
        return ip_network(ip)
    else:
        return ip_address(ip)


def cert_fingerprint(cert: Certificate, hash_algorithm: str = "SHA256") -> str:
    return ":".join(f"{b:02X}" for b in cert.fingerprint(getattr(hashes, hash_algorithm.upper())()))


def cert_is_signed_by_ca(cert: Certificate, ca_cert: Certificate) -> bool:
    try:
        public_key = ca_cert.public_key()
        signature_hash_algorithm = cert.signature_hash_algorithm
        assert isinstance(public_key, RSAPublicKey)
        assert isinstance(signature_hash_algorithm, hashes.HashAlgorithm)
        public_key.verify(
            cert.signature,
            cert.tbs_certificate_bytes,
            padding.PKCS1v15(),
            signature_hash_algorithm,
        )
        return True
    except InvalidSignature:
        return False


def x5t_any(cert: Certificate, hash_algorithm: str) -> str:
    public_bytes = cert.public_bytes(serialization.Encoding.DER)
    hash_instance = hashes.Hash(getattr(hashes, hash_algorithm.upper())(), default_backend())
    hash_instance.update(public_bytes)
    thumbprint = hash_instance.finalize()
    return base64.urlsafe_b64encode(thumbprint).rstrip(b"=").decode("utf-8")


def x5t(cert: Certificate) -> str:
    return x5t_any(cert, "SHA1")


def x5t_s256(cert: Certificate) -> str:
    return x5t_any(cert, "SHA256")



import json
from typing import Optional, Iterator, Dict, Any
from urllib.parse import urlencode

import requests

from resotolib.args import ArgumentParser
from resotolib.baseresources import EdgeType
from resotolib.config import Config
from resotolib.core import resotocore
from resotolib.core.ca import TLSData
from resotolib.core.model_export import node_from_dict, node_to_dict
from resotolib.graph import Graph, sanitize
from resotolib.jwt import encode_jwt_to_headers
from resotolib.logger import log
from resotolib.types import Json


class CoreGraph:
    def __init__(
        self,
        base_uri: Optional[str] = None,
        graph: Optional[str] = None,
        tls_data: Optional[TLSData] = None,
    ) -> None:
        if base_uri is None:
            self.base_uri = resotocore.http_uri
        else:
            self.base_uri = base_uri.strip("/")
        if graph is None:
            self.graph_name = Config.resotoworker.graph
        else:
            self.graph_name = graph
        self.verify = None
        if tls_data:
            self.verify = tls_data.ca_cert_path
        self.graph_uri = f"{self.base_uri}/graph/{self.graph_name}"
        self.search_uri = f"{self.graph_uri}/search/graph"

    def execute(self, command: str) -> Iterator[Json]:
        log.debug(f"Executing command: {command}")
        headers = {"Accept": "application/x-ndjson", "Content-Type": "text/plain"}
        execute_endpoint = f"{self.base_uri}/cli/execute"
        if self.graph_name:
            query_string = urlencode({"graph": self.graph_name})
            execute_endpoint += f"?{query_string}"
        return self.post(execute_endpoint, command, headers, verify=self.verify)

    def search(self, search: str, edge_type: Optional[EdgeType] = None, section: str = "reported") -> Iterator[Json]:
        log.debug(f"Sending search {search}")
        headers = {"Accept": "application/x-ndjson"}
        search_endpoint = self.search_uri
        query = {"section": section}
        if edge_type is not None:
            query["edge_type"] = edge_type.value
        query_string = urlencode(query)
        search_endpoint += f"?{query_string}"
        return self.post(search_endpoint, search, headers, verify=self.verify)

    @staticmethod
    def post(uri: str, data: str, headers: Dict[str, str], verify: Optional[str] = None) -> Iterator[Json]:
        if getattr(ArgumentParser.args, "psk", None):
            encode_jwt_to_headers(headers, {}, ArgumentParser.args.psk)
        r = requests.post(uri, data=data, headers=headers, stream=True, verify=verify)
        if r.status_code != 200:
            log.error(r.content.decode())
            raise RuntimeError(f"Failed to search graph: {r.content.decode()}")
        for line in r.iter_lines():
            if not line:
                continue
            try:
                response: Json = json.loads(line.decode("utf-8"))
                yield response
            except TypeError as e:
                log.error(e)
                continue

    def graph(self, search: str) -> Graph:
        def process_data_line(data: Json, graph: Graph) -> None:
            """Process a single line of resotocore graph data"""

            if data.get("type") == "node":
                node_id = data.get("id")
                node = node_from_dict(data)
                node_mapping[node_id] = node
                log.debug(f"Adding node {node} to the graph")
                graph.add_node(node)
                if node.kind == "graph_root":
                    log.debug(f"Setting graph root {node}")
                    graph.root = node
            elif data.get("type") == "edge":
                node_from = data.get("from")
                node_to = data.get("to")
                edge_type = EdgeType.from_value(data.get("edge_type"))
                if node_from not in node_mapping or node_to not in node_mapping:
                    raise ValueError(f"One of {node_from} -> {node_to} unknown")
                graph.add_edge(node_mapping[node_from], node_mapping[node_to], edge_type=edge_type)

        graph = Graph()
        node_mapping: Dict[Any, Any] = {}
        for data in self.search(search):
            try:
                process_data_line(data, graph)
            except ValueError as e:
                log.error(e)
                continue
        sanitize(graph)
        return graph

    def patch_nodes(self, graph: Graph) -> None:
        headers = {"Content-Type": "application/x-ndjson"}
        if getattr(ArgumentParser.args, "psk", None):
            encode_jwt_to_headers(headers, {}, ArgumentParser.args.psk)

        r = requests.patch(
            f"{self.graph_uri}/nodes",
            data=GraphChangeIterator(graph),
            headers=headers,
            verify=self.verify,
        )
        if r.status_code != 200:
            err = r.content.decode("utf-8")
            log.error(err)
            raise RuntimeError(f"Failed to patch nodes: {err}")


class GraphChangeIterator:
    def __init__(self, graph: Graph) -> None:
        self.graph = graph

    def __iter__(self) -> Iterator[bytes]:
        for node in self.graph.nodes:
            if not node.changes.changed:
                continue
            node_dict = node_to_dict(node, changes_only=True)
            node_json = json.dumps(node_dict) + "\n"
            log.debug(f"Updating node {node_dict}")
            yield node_json.encode()



from __future__ import annotations

import concurrent
from collections import defaultdict, deque
from concurrent.futures import Executor, Future
from functools import reduce
from threading import Event, Lock
from typing import Any, Callable, Deque, Dict, List, Tuple, TypeVar, Optional

from attr import field
from attrs import define

from resotolib.logger import log

T = TypeVar("T")


class CancelOnFirstError(Exception):
    pass


class GatherFutures:
    def __init__(self, futures: List[Future[Any]]) -> None:
        self._futures = futures
        self._lock = Lock()
        self._to_wait = len(futures)
        self._when_done: Future[None] = Future()
        for future in futures:
            future.add_done_callback(self._on_future_done)

    def _on_future_done(self, _: Future[Any]) -> None:
        with self._lock:
            self._to_wait -= 1
            if self._to_wait == 0:
                self._when_done.set_result(None)

    @staticmethod
    def all(futures: List[Future[Any]]) -> Future[None]:
        return GatherFutures(futures)._when_done


@define
class ExecutorQueueTask:
    key: Any
    fn: Callable[..., T]
    args: Tuple[Any, ...]
    kwargs: Dict[str, Any]
    future: Future[Any]

    def __call__(self) -> T:  # type: ignore
        try:
            result: T = self.fn(*self.args, **self.kwargs)
            self.future.set_result(result)
            return result
        except Exception as e:
            self.future.set_exception(e)
            raise


@define
class ExecutorQueue:
    """
    Use an underlying executor to perform work in parallel, but limit the number of tasks per key.
    If fail_on_first_exception_in_group is True, then the first exception in a group
    will not execute any more tasks in the same group.
    """

    executor: Executor
    name: str
    tasks_per_key: Optional[Callable[[str], int]] = None
    fail_on_first_exception_in_group: bool = False
    _tasks_lock: Lock = Lock()
    _tasks: Dict[str, Deque[ExecutorQueueTask]] = field(factory=lambda: defaultdict(deque))
    _in_progress: Dict[str, int] = field(factory=lambda: defaultdict(int))
    _futures: List[Future[Any]] = field(factory=list)
    _exceptions: Dict[Any, Exception] = field(factory=dict)
    _task_finished: Event = Event()

    def submit_work(self, key: Any, fn: Callable[..., T], *args: Any, **kwargs: Any) -> Future[T]:
        future = Future[T]()
        task = ExecutorQueueTask(key=key, fn=fn, args=args, kwargs=kwargs, future=future)
        self.__append_work(task)
        return future

    def __append_work(self, task: ExecutorQueueTask) -> None:
        with self._tasks_lock:
            self._tasks[task.key].appendleft(task)
            self.__check_queue(task.key)

    def __check_queue(self, key: Any) -> None:
        # note: this method is not thread safe, it should only be called from within a lock
        in_progress = self._in_progress[key]
        tasks = self._tasks[key]

        if self.fail_on_first_exception_in_group and self._exceptions.get(key) is not None:
            # Fail all tasks in this group
            ex = CancelOnFirstError("Exception happened in another thread. Do not start work.")
            for task in tasks:
                task.future.set_exception(ex)
            # Clear the queue, so we don't execute them
            # Clear the queue, so we don't execute them
            tasks.clear()

        if ((self.tasks_per_key is None) or (in_progress < self.tasks_per_key(key))) and tasks:
            task = tasks.pop()
            self._in_progress[key] += 1
            self.__perform_task(task)

    def __perform_task(self, task: ExecutorQueueTask) -> Future[T]:
        def only_start_when_no_error() -> T:
            # in case of exception let's fail fast and do not execute the function
            if self._exceptions.get(task.key) is None:
                try:
                    return task()
                except Exception as e:
                    # only store the first exception if we should fail on first future
                    if self._exceptions.get(task.key) is None:
                        self._exceptions[task.key] = e
                    raise e
            else:
                raise CancelOnFirstError(
                    "Exception happened in another thread. Do not start work."
                ) from self._exceptions[task.key]

        def execute() -> T:
            try:
                return only_start_when_no_error() if self.fail_on_first_exception_in_group else task()
            finally:
                with self._tasks_lock:
                    self._in_progress[task.key] -= 1
                    self._task_finished.set()
                    self.__check_queue(task.key)

        future = self.executor.submit(execute)

        self._futures.append(future)
        return future

    def wait_for_submitted_work(self) -> None:
        # wait until all futures are complete
        to_wait = []

        # step 1: wait until all tasks are committed to the executor
        while True:
            with self._tasks_lock:
                ip = reduce(lambda x, y: x + y, self._in_progress.values(), 0)
                if ip == 0:
                    to_wait = self._futures
                    self._futures = []
                    break
                else:
                    # safe inside the lock. clear this event and check when next task is done
                    self._task_finished.clear()
            self._task_finished.wait()

        # step 2: wait for all tasks to complete
        for future in concurrent.futures.as_completed(to_wait):
            try:
                future.result()
            except CancelOnFirstError:
                pass
            except Exception as ex:
                log.exception(f"Unhandled exception in {self.name}: {ex}")
                raise



import pytest
import datetime
from resotolib.jwt import (
    decode_jwt_from_header_value,
    decode_jwt_from_headers,
    encode_jwt,
    decode_jwt,
    encode_jwt_to_headers,
    create_jwk_dict,
)
from jwt import (
    InvalidSignatureError,
    ExpiredSignatureError,
    ImmatureSignatureError,
    MissingRequiredClaimError,
    get_unverified_header,
)
from resotolib.x509 import (
    bootstrap_ca,
    gen_rsa_key,
    gen_csr,
    sign_csr,
    cert_is_signed_by_ca,
    x5t_s256,
)


def test_jwt_psk():
    psk = "somesecret"
    payload = {"Hello": "World"}
    j1 = encode_jwt(payload, psk, expire_in=-1)
    j2 = encode_jwt(payload, psk, expire_in=-1)
    past_date = datetime.datetime.now() - datetime.timedelta(30)
    future_date = datetime.datetime.now() + datetime.timedelta(30)

    expired_payload = {"Hello": "World", "exp": past_date}
    future_payload = {"Hello": "World", "exp": future_date}
    nbf_payload = {"Hello": "World", "nbf": future_date}
    expired_jwt = encode_jwt(expired_payload, psk)
    valid_jwt = encode_jwt(future_payload, psk)
    not_yet_valid_jwt = encode_jwt(nbf_payload, psk)

    h1 = encode_jwt_to_headers({}, payload, psk, expire_in=-1)

    assert j1 != j2
    assert decode_jwt(j1, psk) == payload
    assert decode_jwt(j2, psk) == payload
    assert decode_jwt_from_headers(h1, psk) == payload
    assert decode_jwt_from_headers({}, psk) is None
    assert decode_jwt_from_header_value("", psk) is None
    with pytest.raises(InvalidSignatureError) as e:
        decode_jwt(j1, "wrongpsk")
    assert str(e.value) == "Signature verification failed"
    assert decode_jwt(expired_jwt, psk, options={"verify_exp": False}).get("Hello") == "World"
    with pytest.raises(ExpiredSignatureError) as e:
        decode_jwt(expired_jwt, psk)
    assert str(e.value) == "Signature has expired"
    assert decode_jwt(valid_jwt, psk).get("Hello") == "World"
    with pytest.raises(ImmatureSignatureError) as e:
        decode_jwt(not_yet_valid_jwt, psk)
    assert str(e.value).startswith("The token is not yet valid")
    with pytest.raises(MissingRequiredClaimError) as e:
        decode_jwt(j1, psk, options={"require": ["exp"]})
    assert str(e.value) == 'Token is missing the "exp" claim'


def test_jwt_pki():
    ca_key, ca_cert = bootstrap_ca()
    cert_key = gen_rsa_key()
    cert_crt = sign_csr(gen_csr(cert_key), ca_key, ca_cert)
    payload = {"Hello": "World"}
    jwt = encode_jwt(payload, cert_key, expire_in=-1)
    assert cert_is_signed_by_ca(cert_crt, ca_cert)
    assert decode_jwt(jwt, cert_crt) == payload

    jwt = encode_jwt(payload, cert_key, cert=cert_crt)
    assert decode_jwt(jwt, cert_crt).get("Hello") == "World"
    assert get_unverified_header(jwt).get("x5t#S256") == x5t_s256(cert_crt)


def test_jwk() -> None:
    ca_key, ca_cert = bootstrap_ca()
    cert_key = gen_rsa_key()
    cert_crt = sign_csr(gen_csr(cert_key), ca_key, ca_cert)
    cert_cwk = create_jwk_dict(cert_crt)
    # has 9 entries
    assert len(cert_cwk) == 9
    # creating the jwk from the same cert, creates the same key data
    assert cert_cwk == create_jwk_dict(cert_crt)
    # check on specific values
    assert cert_cwk["alg"] == "sha256WithRSAEncryption"
    assert cert_cwk["kty"] == "RSA"
    assert cert_cwk["use"] == "sig"
    assert cert_cwk["x5t#S256"] == x5t_s256(cert_crt)
    assert cert_cwk["kid"] == cert_cwk["x5t#S256"]



import re
from collections import defaultdict
from datetime import datetime, date, timedelta
from enum import Enum
from functools import lru_cache, reduce
from pydoc import locate
from typing import List, MutableSet, Union, Tuple, Dict, Set, Any, TypeVar, Type, Optional
from typing import get_args, get_origin

import attrs
from attr import resolve_types
from attrs import Attribute

from resotolib.baseresources import BaseResource
from resotolib.json import from_json
from resotolib.types import Json
from resotolib.utils import type_str

property_metadata_to_strip = ["restart_required", "description", "required", "kind"]


# List[X] -> list, list -> list
def optional_origin(clazz: Type[Any]) -> Type[Any]:
    maybe_optional = get_args(clazz)[0] if is_optional(clazz) else clazz
    origin = get_origin(maybe_optional)
    return origin if origin else maybe_optional  # type: ignore


# Optional[x] -> true
def is_optional(clazz: Union[type, Tuple[Any]]) -> bool:
    args = get_args(clazz)
    return get_origin(clazz) is Union and type(None) in args and len(args) == 2


# List[x] -> true, list -> true
def is_collection(clazz: type) -> bool:
    return optional_origin(clazz) in [list, set, tuple]


# Dict[x,y] -> true, dict -> true
def is_dict(clazz: type) -> bool:
    return optional_origin(clazz) in [dict]


# either enum or optional enum
def is_enum(clazz: type) -> bool:
    origin = optional_origin(clazz)
    return isinstance(origin, type) and issubclass(origin, Enum)


# List[X] -> X, list -> object
def type_arg(clazz: type) -> type:
    maybe_optional = get_args(clazz)[0] if is_optional(clazz) else clazz
    args = get_args(maybe_optional)
    return args[0] if args and len(args) == 1 else object  # type: ignore


# Dict[X,Y] -> (X,Y), dict -> (object, object)
def dict_types(clazz: type) -> Tuple[type, type]:
    maybe_optional = get_args(clazz)[0] if is_optional(clazz) else clazz
    args = get_args(maybe_optional)
    return (args[0], args[1]) if args and len(args) == 2 else (object, object)


# walk class hierarchy, as well as all properties to find transitive data classes
def transitive_classes(classes: Set[type], walk_subclasses: bool = True) -> Set[type]:
    all_classes: MutableSet[type] = set()

    def check(to_check: type) -> None:
        clazz = optional_origin(to_check)
        if clazz in all_classes:
            pass
        elif is_dict(clazz):
            key_type, value_type = dict_types(to_check)
            check(key_type)
            check(value_type)
        elif is_collection(clazz):
            check(type_arg(to_check))
        elif attrs.has(clazz):
            resolve_types(clazz)
            all_classes.add(clazz)
            for mro_clazz in clazz.mro()[1:]:
                check(mro_clazz)
            if walk_subclasses:
                for subclass in clazz.__subclasses__():
                    check(subclass)
            for field in attrs.fields(clazz):
                check(field.type)
        elif is_enum(clazz):
            all_classes.add(clazz)

    for c in classes:
        check(c)

    return set(all_classes)


lookup: Dict[type, str] = {
    str: "string",
    int: "int64",
    bool: "boolean",
    float: "double",
    datetime: "datetime",
    date: "date",
    timedelta: "duration",
}
simple_type = tuple(lookup.keys())


# Model name from the internal python class name
def model_name(clazz: Union[type, Tuple[Any], None]) -> str:
    if clazz is None:
        return "any"
    to_check = get_args(clazz)[0] if is_optional(clazz) else clazz
    if is_collection(to_check):
        return f"{model_name(type_arg(to_check))}[]"
    elif is_dict(to_check):
        key_type, value_type = dict_types(to_check)
        return f"dictionary[{model_name(key_type)}, {model_name(value_type)}]"
    elif is_enum(to_check):
        # camel case to snake case
        return re.sub(r"(?<!^)(?=[A-Z])", "_", to_check.__name__).lower()
    elif get_origin(to_check) == Union:
        # this is a union of different types other than none.
        # since union types are not supported, we fallback to any here
        return "any"
    elif isinstance(to_check, TypeVar):
        return model_name(get_args(to_check))
    elif isinstance(to_check, type) and issubclass(to_check, simple_type):
        return lookup[to_check]
    elif attrs.has(to_check):
        name = getattr(to_check, "kind", None)
        if not name:
            raise AttributeError(f"dataclass {to_check} need to define a ClassVar kind!")
        return name  # type: ignore
    else:
        return "any"


# define if a field should be exported or not.
# Use python default: hide props starting with underscore.
def should_export(field: Attribute) -> bool:  # type: ignore
    return not field.name.startswith("_")


def dataclasses_to_resotocore_model(
    classes: Set[Type[Any]],
    *,
    allow_unknown_props: bool = False,
    aggregate_root: Optional[Type[Any]] = None,
    walk_subclasses: bool = True,
    use_optional_as_required: bool = False,
    with_description: bool = True,
) -> List[Json]:
    """
    Analyze all transitive dataclasses and create the model
    definition as understood by resotocore.
    A plain python dataclass defines the model structure and
    should be used to create json in the same format.

    :param classes: all dataclasses to analyze.
    :param allow_unknown_props: allow properties in json that are not defined in the model.
    :param aggregate_root: if a type is a subtype of this type, it will be considered an aggregate root.
    :param walk_subclasses: if true, all subclasses of the given classes will be analyzed as well.
    :param use_optional_as_required: if true, all non-optional fields will be considered required.
    :param with_description: if true, include the description for classes and properties.
    :return: the model definition in the resotocore json format.
    """

    def prop(field: Attribute) -> List[Json]:  # type: ignore
        # the field itself can define the type via a type hint
        # this is useful for int and float in python where the representation can not be
        # detected by the type itself. Example: int32/int64 or float/double
        # If not defined, we fall back to the largest container: int64 and double
        name = field.name
        meta = field.metadata.copy()
        kind = meta.pop("type_hint", model_name(field.type))
        desc = meta.pop("description", None)
        desc = desc if with_description else None
        required = meta.pop("required", use_optional_as_required and not is_optional(field.type))  # type: ignore
        synthetic = meta.pop("synthetic", None)
        synthetic = synthetic if synthetic else {}
        for ps in property_metadata_to_strip:
            meta.pop(ps, None)

        def json(
            name: str,
            kind_str: str,
            required: bool,
            description: Optional[str],
            meta: Optional[Dict[str, str]],
            **kwargs: Any,
        ) -> Json:
            js = {"name": name, "kind": kind_str, "required": required, "description": description, **kwargs}
            if meta:
                js["metadata"] = meta
            return js

        synthetics = [
            json(
                synth_prop,
                synth_trafo,
                False,
                None,
                None,
                synthetic={"path": [name]},
            )
            for synth_prop, synth_trafo in synthetic.items()
        ]

        return [json(name, kind, required, desc, meta)] + synthetics

    for cls in classes:
        if attrs.has(cls):
            resolve_types(cls)
    model: List[Json] = []
    all_classes = transitive_classes(classes, walk_subclasses)

    # type edge_type -> list of types
    successors: Dict[str, Dict[str, List[str]]] = defaultdict(lambda: defaultdict(list))

    for clazz in all_classes:
        if name := getattr(clazz, "kind", None):
            # backwards compatibility: still look for successor kinds variable. deprecated.
            if succs := getattr(clazz, "successor_kinds", None):
                for edge_type, values in succs.items():
                    successors[name][edge_type].extend(values)
            if refs := getattr(clazz, "reference_kinds", None):
                if succs := refs.get("successors", None):
                    for edge_type, values in succs.items():
                        successors[name][edge_type].extend(values)
                if preds := refs.get("predecessors", None):
                    for edge_type, values in preds.items():
                        for value in values:
                            successors[value][edge_type].append(name)
    successors = {k: {ik: list(set(iv)) for ik, iv in v.items()} for k, v in successors.items()}

    def export_data_class(clazz: type) -> None:
        bases = [base for base in clazz.__bases__ if attrs.has(base)]
        base_names = [model_name(base) for base in bases]
        base_props: Set[Attribute] = reduce(lambda result, base: result | set(attrs.fields(base)), bases, set())  # type: ignore # noqa: E501
        props = [
            p for field in attrs.fields(clazz) if field not in base_props and should_export(field) for p in prop(field)
        ]
        root = any(sup == aggregate_root for sup in clazz.mro()) if aggregate_root else True
        kind = model_name(clazz)
        metadata: Json = {}
        if (m := getattr(clazz, "metadata", None)) and isinstance(m, dict):
            metadata = m.copy()
        if (s := clazz.__dict__.get("kind_display", None)) and isinstance(s, str):
            metadata["name"] = s
        if with_description and (s := clazz.__dict__.get("kind_description", None)) and isinstance(s, str):
            metadata["description"] = s

        model.append(
            {
                "fqn": kind,
                "bases": base_names,
                "properties": props,
                "allow_unknown_props": allow_unknown_props,
                "successor_kinds": successors.get(kind, None),
                "aggregate_root": root,
                "metadata": metadata or None,
            }
        )

    def export_enum(clazz: Type[Enum]) -> None:
        # Only allow string based enumerations.
        def literal_name(en: Enum) -> str:
            if isinstance(en.value, str):
                return en.value
            else:
                raise AttributeError(f"Enumeration {clazz} does not use string as values!")

        enum_values = [literal_name(literal) for literal in clazz]
        model.append({"fqn": model_name(clazz), "runtime_kind": "string", "enum": enum_values})

    for cls in all_classes:
        if attrs.has(cls):
            resolve_types(cls)
            export_data_class(cls)
        elif is_enum(cls):
            export_enum(cls)
        else:
            raise AttributeError(f"Don't know how to handle: {cls}")
    return model


# Use this model exporter, if a dynamic object is exported
# with given name and properties.
def dynamic_object_to_resotocore_model(
    name: str, properties: Dict[str, type], aggregate_root: bool = True, traverse_dependant: bool = True
) -> List[Json]:
    dependant = dataclasses_to_resotocore_model(set(properties.values())) if traverse_dependant else []
    # append definition for top level object
    dependant.append(
        {
            "fqn": name,
            "bases": [],
            "aggregate_root": aggregate_root,
            "properties": [
                {"name": prop_name, "kind": model_name(prop_type), "required": False}
                for prop_name, prop_type in properties.items()
            ],
            "metadata": {"dynamic": True},
        }
    )
    return dependant


def get_node_attributes(node: BaseResource) -> Json:
    if not hasattr(node, "to_json"):
        raise ValueError(f"Node {node} has no to_json() method!")
    result = node.to_json()
    result["kind"] = node.kind
    return result


def node_to_dict(node: BaseResource, changes_only: bool = False, include_revision: bool = False) -> Json:
    node_dict: Json = {"id": node._resotocore_id if node._resotocore_id else node.chksum}
    if changes_only:
        node_dict.update(node.changes.get())
    else:
        node_dict.update(
            {
                "reported": get_node_attributes(node),
                "metadata": {
                    "python_type": type_str(node),
                    "cleaned": node.cleaned,
                    "phantom": node.phantom,
                    "protected": node.protected,
                    **node._metadata,
                },
                "usage": node._resource_usage,
            }
        )
        if node.clean:
            node_dict.update(
                {
                    "desired": {
                        "clean": node.clean,
                    }
                }
            )
    if include_revision and node._resotocore_revision:
        node_dict.update(
            {
                "revision": node._resotocore_revision,
            }
        )
    return node_dict


@lru_cache(maxsize=None)
def locate_python_type(python_type: str) -> Any:
    cls: Type[Any] = locate(python_type)  # type: ignore
    if attrs.has(cls):
        attrs.resolve_types(cls)
    return cls


def node_from_dict(node_data: Json, include_select_ancestors: bool = False) -> BaseResource:
    """Create a resource from resotocore graph node data

    If include_select_ancestors is True, the resource will be created with
    ancestors cloud, account, region and zone passed directly to the constructor.
    This is useful for when the node is being created for standalone use instead
    of in the context of a graph. E.g. during tagging of an individual resource
    where the tagger needs to know which cloud and account a resource came from.
    """
    node_data_reported = node_data.get("reported", {})
    if node_data_reported is None:
        node_data_reported = {}
    node_data_desired = node_data.get("desired", {})
    if node_data_desired is None:
        node_data_desired = {}
    node_data_metadata = node_data.get("metadata", {})
    if node_data_metadata is None:
        node_data_metadata = {}
    node_data_ancestors = node_data.get("ancestors", {})
    if node_data_ancestors is None:
        node_data_ancestors = {}

    new_node_data = dict(node_data_reported)
    if "kind" in new_node_data:
        del new_node_data["kind"]

    python_type = node_data_metadata.get("python_type", "NoneExisting")
    node_type = locate_python_type(python_type)
    if node_type is None:
        raise ValueError(f"Do not know how to handle {node_data_reported}")

    cleanup_node_field_types(node_type, new_node_data)
    ancestors = {}
    if include_select_ancestors:
        for ancestor in ("cloud", "account", "region", "zone"):
            if node_data_ancestors.get(ancestor):
                ancestors[f"_{ancestor}"] = node_from_dict(
                    {
                        "reported": node_data_ancestors[ancestor].get("reported", {}),
                        "metadata": node_data_ancestors[ancestor].get("metadata", {}),
                    }
                )
    new_node_data.update(
        {
            "_resotocore_id": node_data.get("id"),
            "_resotocore_revision": node_data.get("revision"),
            "_resotocore_query_tag": node_data_metadata.get("query_tag"),
        }
    )

    node: BaseResource = from_json(new_node_data, node_type)
    for field_name, value in ancestors.items():
        setattr(node, field_name, value)
    node._raise_tags_exceptions = True
    node._protected = node_data_metadata.get("protected", False)
    node._cleaned = node_data_metadata.get("cleaned", False)
    node._clean = node_data_desired.get("clean", False)
    return node


def cleanup_node_field_types(node_type: BaseResource, node_data_reported: Json) -> None:
    valid_fields = set(field.name for field in attrs.fields(node_type))  # type: ignore
    for field_name in list(node_data_reported.keys()):
        if field_name not in valid_fields:
            del node_data_reported[field_name]



from networkx import DiGraph
from pytest import fixture

from resotolib.graph.graph_extensions import dependent_node_iterator


@fixture
def graph() -> DiGraph:
    g = DiGraph()
    for i in range(1, 14):
        g.add_node(i)
    g.add_edges_from([(1, 2), (1, 3), (2, 3)])  # island 1
    g.add_edges_from([(4, 5), (4, 6), (6, 7)])  # island 2
    g.add_edges_from([(8, 9), (9, 10), (9, 11), (8, 12), (12, 11), (12, 13)])  # island 3
    return g


def test_reversed_directed_traversal(graph: DiGraph):
    result = list(dependent_node_iterator(graph))
    assert len(result) == 3  # 3 steps to complete
    assert result == [
        [3, 5, 7, 10, 11, 13],  # step 1
        [2, 6, 9, 12],  # step 2
        [1, 4, 8],  # step 3
    ]


def test_delete_nodes(graph: DiGraph):
    to_delete = graph.copy()
    for parallel in dependent_node_iterator(graph):
        for node in parallel:
            to_delete.remove_node(node)
    assert len(to_delete.nodes) == 0


def test_empty_graph():
    assert list(dependent_node_iterator(DiGraph())) == []



from __future__ import annotations

import json
import queue
import threading
import time
from typing import Callable, Dict, Optional, List, Any
from urllib.parse import urlunsplit, urlsplit

from attrs import define, field
from websocket import WebSocketApp, WebSocket  # type: ignore

from resotolib.args import ArgumentParser
from resotolib.baseresources import BaseResource
from resotolib.config import current_config
from resotolib.core.ca import TLSData
from resotolib.core.custom_command import CommandDefinition
from resotolib.core.model_export import node_to_dict, node_from_dict
from resotolib.event import EventType, remove_event_listener, add_event_listener, Event
from resotolib.json import to_json_str
from resotolib.jwt import encode_jwt_to_headers
from resotolib.logger import log
from resotolib.types import Json, JsonElement


@define
class CoreTaskResult:
    task_id: str
    data: JsonElement = None
    error: Optional[str] = None

    def to_json(self) -> Json:
        if self.error:
            return {"task_id": self.task_id, "result": "error", "error": self.error}
        else:
            return {"task_id": self.task_id, "result": "done", "data": self.data}


@define
class CoreTaskHandler:
    name: str
    info: str
    description: str
    handler: Callable[[Json], JsonElement]
    expect_node_result: bool = False
    filter: Dict[str, List[str]] = field(factory=dict)
    allowed_on_kind: Optional[str] = None
    args_description: Dict[str, str] = field(factory=dict)

    def execute(self, message: Json) -> CoreTaskResult:
        task_id = message["task_id"]  # fail if there is no task_id
        try:
            task_data: Json = message.get("data", {})
            result = self.handler(task_data)
            return CoreTaskResult(task_id=task_id, data=result)
        except Exception as e:
            log.debug(f"Error while executing task {self.name}: {e}", exc_info=True)
            return CoreTaskResult(task_id=task_id, error=str(e))

    def matches(self, js: Json) -> bool:
        attrs: Json = js.get("attrs", {})
        if js.get("task_name") != self.name or not isinstance(attrs, dict):
            return False
        return all((attrs.get(n) in f) for n, f in self.filter.items())

    def core_json(self) -> Json:
        return {
            "name": self.name,
            "info": self.info,
            "description": self.description,
            "filter": self.filter,
            "expect_node_result": self.expect_node_result,
            "args_description": self.args_description,
            "allowed_on_kind": self.allowed_on_kind,
        }

    @staticmethod
    def from_definition(target: Any, wtd: CommandDefinition) -> CoreTaskHandler:
        def handle_message(message: Json) -> JsonElement:
            node_data = message.get("node", {})
            args = message.get("args", [])
            return wtd.fn(target, current_config(), node_data, args)  # type: ignore

        def handle_resource(message: Json) -> JsonElement:
            node_data = message.get("node", {})
            node = node_from_dict(node_data, include_select_ancestors=True) if node_data else None
            args = message.get("args", [])
            result = wtd.fn(target, current_config(), node, args)
            # expect either a base resource or json element as result
            if isinstance(result, BaseResource):
                return node_to_dict(result)
            else:
                return result  # type: ignore

        to_call = handle_resource if wtd.expect_resource else handle_message
        return CoreTaskHandler(
            name=wtd.name,
            info=wtd.info,
            args_description=wtd.args_description,
            description=wtd.description,
            filter=wtd.filter,
            expect_node_result=wtd.expect_node_result,
            allowed_on_kind=wtd.allowed_on_kind,
            handler=to_call,
        )


class CoreTasks(threading.Thread):
    def __init__(
        self,
        identifier: str,
        resotocore_ws_uri: str,
        task_handler: List[CoreTaskHandler],
        max_workers: int = 20,
        tls_data: Optional[TLSData] = None,
    ) -> None:
        super().__init__()
        self.identifier = identifier
        self.resotocore_ws_uri = resotocore_ws_uri
        self.task_handler = task_handler
        self.max_workers = max_workers
        self.tls_data = tls_data
        self.ws: Optional[WebSocketApp] = None
        self.shutdown_event = threading.Event()
        self.queue: queue.Queue[Json] = queue.Queue()
        self.__connected = False

    def connected(self) -> bool:
        return self.__connected

    def __del__(self) -> None:
        remove_event_listener(EventType.SHUTDOWN, self.shutdown)

    def run(self) -> None:
        self.name = self.identifier
        add_event_listener(EventType.SHUTDOWN, self.shutdown)

        for i in range(self.max_workers):
            threading.Thread(target=self.worker, daemon=True, name=f"worker-{i}").start()

        while not self.shutdown_event.is_set():
            log.debug("Connecting to resotocore task queue")
            try:
                self.connect()
            except Exception as e:
                log.error(e)
            time.sleep(1)

    def worker(self) -> None:
        while not self.shutdown_event.is_set():
            message = self.queue.get()
            log.debug(f"{self.identifier} received: {message}")
            for handler in self.task_handler:
                if handler.matches(message):
                    try:
                        result = handler.execute(message)
                        if self.ws:
                            log.debug(f"Sending reply {result.to_json()}")
                            self.ws.send(json.dumps(result.to_json()))
                    except Exception as ex:
                        log.exception(f"Something went wrong while processing {message}")
                        if (task_id := message.get("task_id")) and self.ws is not None:
                            self.ws.send(to_json_str(CoreTaskResult(task_id, error=str(ex)).to_json()))
                    break
            self.queue.task_done()

    def connect(self) -> None:
        resotocore_ws_uri_split = urlsplit(self.resotocore_ws_uri)
        scheme = resotocore_ws_uri_split.scheme
        netloc = resotocore_ws_uri_split.netloc
        path = resotocore_ws_uri_split.path + "/work/queue"
        ws_uri = urlunsplit((scheme, netloc, path, "", ""))

        log.debug(f"{self.identifier} connecting to {ws_uri}")
        headers: Dict[str, str] = {}
        if getattr(ArgumentParser.args, "psk", None):
            encode_jwt_to_headers(headers, {}, ArgumentParser.args.psk)
        self.ws = WebSocketApp(
            ws_uri,
            header=headers,
            on_open=self.on_open,
            on_message=self.on_message,
            on_error=self.on_error,
            on_close=self.on_close,
            on_ping=self.on_ping,
            on_pong=self.on_pong,
        )
        sslopt: Dict[Any, Any] = {}
        if self.tls_data:
            sslopt = {"ca_certs": self.tls_data.ca_cert_path}
        self.ws.run_forever(sslopt=sslopt, ping_interval=20, ping_timeout=10, ping_payload="ping")

    def shutdown(self, _: Optional[Event] = None) -> None:
        log.debug("Received shutdown event - shutting down resotocore task queue listener")
        self.shutdown_event.set()
        if self.ws:
            self.ws.close()

    def on_message(self, _: WebSocket, message: str) -> None:
        try:
            jsom_message: Json = json.loads(message)
        except json.JSONDecodeError:
            log.exception(f"Unable to decode received message {message}")
            return
        self.queue.put(jsom_message)

    def on_error(self, _: WebSocket, e: Exception) -> None:
        log.debug(f"{self.identifier} event bus error: {e!r}")

    def on_close(self, _: WebSocket, close_status_code: int, close_msg: str) -> None:
        self.__connected = False
        log.debug(f"{self.identifier} disconnected from resotocore task queue")

    def on_open(self, ws: WebSocket) -> None:
        self.__connected = True
        log.debug(f"{self.identifier} connected to resotocore, register at task queue")
        # when we are connected, we register at the task queue
        # by sending all task handler definitions
        ws.send(json.dumps([handler.core_json() for handler in self.task_handler]))

    def on_ping(self, _: WebSocket, message: str) -> None:
        log.debug(f"{self.identifier} tasks ping from resotocore message bus")

    def on_pong(self, _: WebSocket, message: str) -> None:
        log.debug(f"{self.identifier} tasks pong from resotocore message bus")



import unittest
import threading
import time
import copy
from datetime import datetime

try:
    from zoneinfo import ZoneInfo
except ImportError:
    from backports.zoneinfo import ZoneInfo
from tempfile import TemporaryDirectory
from resotolib.lock import RWLock
from resotolib.utils import (
    ordinal,
    sha256sum,
    rrdata_as_dict,
    get_local_tzinfo,
    utc_str,
    replace_env_vars,
    merge_json_elements,
    drop_deleted_attributes,
    freeze,
)
from resotolib.baseresources import BaseResource
from resotolib.utils import stdin_generator
from attrs import define
from typing import ClassVar
import pytest
import sys
import tempfile
from contextlib import contextmanager
from frozendict import frozendict


class Writer(threading.Thread):
    def __init__(self, buffer_, rw_lock, init_sleep_time, sleep_time, to_write):
        """
        @param buffer_: common buffer_ shared by the readers and writers
        @type buffer_: list
        @type rw_lock: L{RWLock}
        @param init_sleep_time: sleep time before doing any action
        @type init_sleep_time: C{float}
        @param sleep_time: sleep time while in critical section
        @type sleep_time: C{float}
        @param to_write: data that will be appended to the buffer
        """
        threading.Thread.__init__(self)
        self.__buffer = buffer_
        self.__rw_lock = rw_lock
        self.__init_sleep_time = init_sleep_time
        self.__sleep_time = sleep_time
        self.__to_write = to_write
        self.entry_time = None
        """Time of entry to the critical section"""
        self.exit_time = None
        """Time of exit from the critical section"""

    def run(self):
        time.sleep(self.__init_sleep_time)
        self.__rw_lock.writer_acquire()
        self.entry_time = time.time()
        time.sleep(self.__sleep_time)
        self.__buffer.append(self.__to_write)
        self.exit_time = time.time()
        self.__rw_lock.writer_release()


class Reader(threading.Thread):
    def __init__(self, buffer_, rw_lock, init_sleep_time, sleep_time):
        """
        @param buffer_: common buffer shared by the readers and writers
        @type buffer_: list
        @type rw_lock: L{RWLock}
        @param init_sleep_time: sleep time before doing any action
        @type init_sleep_time: C{float}
        @param sleep_time: sleep time while in critical section
        @type sleep_time: C{float}
        """
        threading.Thread.__init__(self)
        self.__buffer = buffer_
        self.__rw_lock = rw_lock
        self.__init_sleep_time = init_sleep_time
        self.__sleep_time = sleep_time
        self.buffer_read = None
        """a copy of a the buffer read while in critical section"""
        self.entry_time = None
        """Time of entry to the critical section"""
        self.exit_time = None
        """Time of exit from the critical section"""

    def run(self):
        time.sleep(self.__init_sleep_time)
        self.__rw_lock.reader_acquire()
        self.entry_time = time.time()
        time.sleep(self.__sleep_time)
        self.buffer_read = copy.deepcopy(self.__buffer)
        self.exit_time = time.time()
        self.__rw_lock.reader_release()


class RWLockTestCase(unittest.TestCase):
    def test_readers_nonexclusive_access(self):
        (buffer_, rw_lock, threads) = self.__init_variables()

        threads.append(Reader(buffer_, rw_lock, 0, 0))
        threads.append(Writer(buffer_, rw_lock, 0.2, 0.4, 1))
        threads.append(Reader(buffer_, rw_lock, 0.3, 0.3))
        threads.append(Reader(buffer_, rw_lock, 0.5, 0))

        self.__start_and_join_threads(threads)

        # The third reader should enter after the second one but it should
        # exit before the second one exits
        # (i.e. the readers should be in the critical section
        # at the same time)

        self.assertEqual([], threads[0].buffer_read)
        self.assertEqual([1], threads[2].buffer_read)
        self.assertEqual([1], threads[3].buffer_read)
        self.assertTrue(threads[1].exit_time <= threads[2].entry_time)
        self.assertTrue(threads[2].entry_time <= threads[3].entry_time)
        self.assertTrue(threads[3].exit_time < threads[2].exit_time)

    def test_writers_exclusive_access(self):
        (buffer_, rw_lock, threads) = self.__init_variables()

        threads.append(Writer(buffer_, rw_lock, 0, 0.4, 1))
        threads.append(Writer(buffer_, rw_lock, 0.1, 0, 2))
        threads.append(Reader(buffer_, rw_lock, 0.2, 0))

        self.__start_and_join_threads(threads)

        # The second writer should wait for the first one to exit

        self.assertEqual([1, 2], threads[2].buffer_read)
        self.assertTrue(threads[0].exit_time <= threads[1].entry_time)
        self.assertTrue(threads[1].exit_time <= threads[2].exit_time)

    def test_writer_priority(self):
        (buffer_, rw_lock, threads) = self.__init_variables()

        threads.append(Writer(buffer_, rw_lock, 0, 0, 1))
        threads.append(Reader(buffer_, rw_lock, 0.1, 0.4))
        threads.append(Writer(buffer_, rw_lock, 0.2, 0, 2))
        threads.append(Reader(buffer_, rw_lock, 0.3, 0))
        threads.append(Reader(buffer_, rw_lock, 0.3, 0))

        self.__start_and_join_threads(threads)

        # The second writer should go before the second and the third reader

        self.assertEqual([1], threads[1].buffer_read)
        self.assertEqual([1, 2], threads[3].buffer_read)
        self.assertEqual([1, 2], threads[4].buffer_read)
        self.assertTrue(threads[0].exit_time < threads[1].entry_time)
        self.assertTrue(threads[1].exit_time <= threads[2].entry_time)
        self.assertTrue(threads[2].exit_time <= threads[3].entry_time)
        self.assertTrue(threads[2].exit_time <= threads[4].entry_time)

    def test_many_writers_priority(self):
        (buffer_, rw_lock, threads) = self.__init_variables()

        threads.append(Writer(buffer_, rw_lock, 0, 0, 1))
        threads.append(Reader(buffer_, rw_lock, 0.1, 0.6))
        threads.append(Writer(buffer_, rw_lock, 0.2, 0.1, 2))
        threads.append(Reader(buffer_, rw_lock, 0.3, 0))
        threads.append(Reader(buffer_, rw_lock, 0.4, 0))
        threads.append(Writer(buffer_, rw_lock, 0.5, 0.1, 3))

        self.__start_and_join_threads(threads)

        # The two last writers should go first -- after the first reader and
        # before the second and the third reader

        self.assertEqual([1], threads[1].buffer_read)
        self.assertEqual([1, 2, 3], threads[3].buffer_read)
        self.assertEqual([1, 2, 3], threads[4].buffer_read)
        self.assertTrue(threads[0].exit_time < threads[1].entry_time)
        self.assertTrue(threads[1].exit_time <= threads[2].entry_time)
        self.assertTrue(threads[1].exit_time <= threads[5].entry_time)
        self.assertTrue(threads[2].exit_time <= threads[3].entry_time)
        self.assertTrue(threads[2].exit_time <= threads[4].entry_time)
        self.assertTrue(threads[5].exit_time <= threads[3].entry_time)
        self.assertTrue(threads[5].exit_time <= threads[4].entry_time)

    @staticmethod
    def __init_variables():
        buffer_ = []
        rw_lock = RWLock()
        threads = []
        return (buffer_, rw_lock, threads)

    @staticmethod
    def __start_and_join_threads(threads):
        for t in threads:
            t.start()
        for t in threads:
            t.join()


@define(eq=False, slots=False)
class SomeTestResource(BaseResource):
    kind: ClassVar[str] = "some_test_resource"

    def __attrs_post_init__(self) -> None:
        super().__attrs_post_init__()
        self.key = None
        self.value = None

    def update_tag(self, key, value):
        self.key = key
        self.value = value
        return True

    def delete(self, graph) -> bool:
        return False


def test_ordinal():
    assert ordinal(1) == "1st"
    assert ordinal(2) == "2nd"
    assert ordinal(3) == "3rd"
    assert ordinal(4) == "4th"
    assert ordinal(11) == "11th"
    assert ordinal(12) == "12th"
    assert ordinal(13) == "13th"
    assert ordinal(21) == "21st"
    assert ordinal(22) == "22nd"
    assert ordinal(23) == "23rd"


def test_sha256sum():
    test_string = b"Hello World!"
    expected_sha256sum = "7f83b1657ff1fc53b92dc18148a1d65dfc2d4b1fa3d677284addd200126d9069"
    with TemporaryDirectory() as tmp:
        tmp_file = f"{tmp}/testfile"
        with open(tmp_file, "wb") as f:
            f.write(test_string)
        assert sha256sum(tmp_file) == expected_sha256sum


def test_rrdata_as_dict():
    test_soa1 = """ns.icann.org. noc.dns.icann.org. (
        2020080302  ;Serial
        7200        ;Refresh
        3600        ;Retry
        1209600     ;Expire
        3600        ;Negative response caching TTL
)
"""
    test_soa2 = "ns-1620.awsdns-10.co.uk. awsdns-hostmaster.amazon.com. 1 7200 900 1209600 86400"
    test_txt1 = '"Hello World"'
    test_txt2 = '"foo" "bar"'
    test_txt3 = (
        '"v=spf1 a mx ip4:192.168.0.0/24 ip4:192.168.1.0/24 ip4:192.168.1.0/24'
        " ip4:192.168.1.0/24 ip4:192.168.1.0/24 ip4:192.168.1.0/24 ip4:192.168.1.0/24"
        ' ip4:1" "92.168.1.0/24 ip4:192.168.1.0/24 ip4:192.168.1.0/24'
        ' ip4:192.168.1.0/24 ip4:192.168.1.0/24 ip4:192.168.1.0/24 -all"'
    )
    test_txt4 = r'"Hello \" "World " "this is a test" "of TXT"'
    test_mx = "1 ASPMX.L.GOOGLE.com."
    test_a = "10.0.0.1"
    test_aaaa = "2a05:d014:275:cb00:7dff:602c:d0e7:9c4"
    test_ns = "ns-137-b.gandi.net."
    test_srv = "0 10 443 resoto.com."
    res_soa1 = rrdata_as_dict("SOA", test_soa1)
    res_soa2 = rrdata_as_dict("SOA", test_soa2)
    res_txt1 = rrdata_as_dict("TXT", test_txt1)
    res_txt2 = rrdata_as_dict("TXT", test_txt2)
    res_txt3 = rrdata_as_dict("TXT", test_txt3)
    res_txt4 = rrdata_as_dict("TXT", test_txt4)
    res_mx = rrdata_as_dict("MX", test_mx)
    res_a = rrdata_as_dict("A", test_a)
    res_aaaa = rrdata_as_dict("AAAA", test_aaaa)
    res_ns = rrdata_as_dict("NS", test_ns)
    res_srv = rrdata_as_dict("SRV", test_srv)
    assert res_a["record_value"] == test_a
    assert res_aaaa["record_value"] == test_aaaa
    assert res_ns["record_value"] == test_ns
    assert res_mx["record_value"] == "ASPMX.L.GOOGLE.com."
    assert res_mx["record_priority"] == 1
    assert res_txt1["record_value"] == "Hello World"
    assert res_txt2["record_value"] == "foobar"
    assert res_txt3["record_value"] == (
        "v=spf1 a mx ip4:192.168.0.0/24 ip4:192.168.1.0/24 ip4:192.168.1.0/24"
        " ip4:192.168.1.0/24 ip4:192.168.1.0/24 ip4:192.168.1.0/24"
        " ip4:192.168.1.0/24 ip4:192.168.1.0/24 ip4:192.168.1.0/24"
        " ip4:192.168.1.0/24 ip4:192.168.1.0/24 ip4:192.168.1.0/24"
        " ip4:192.168.1.0/24 -all"
    )
    assert res_txt4["record_value"] == r'Hello \" "World this is a testof TXT'
    assert res_srv["record_priority"] == 0
    assert res_srv["record_weight"] == 10
    assert res_srv["record_port"] == 443
    assert res_srv["record_value"] == "resoto.com."
    assert res_soa1["record_mname"] == "ns.icann.org."
    assert res_soa1["record_rname"] == "noc.dns.icann.org."
    assert res_soa1["record_serial"] == 2020080302
    assert res_soa1["record_refresh"] == 7200
    assert res_soa1["record_retry"] == 3600
    assert res_soa1["record_expire"] == 1209600
    assert res_soa1["record_minimum"] == 3600
    assert res_soa2["record_mname"] == "ns-1620.awsdns-10.co.uk."
    assert res_soa2["record_rname"] == "awsdns-hostmaster.amazon.com."
    assert res_soa2["record_serial"] == 1
    assert res_soa2["record_refresh"] == 7200
    assert res_soa2["record_retry"] == 900
    assert res_soa2["record_expire"] == 1209600
    assert res_soa2["record_minimum"] == 86400


def test_get_local_tzinfo():
    tz = get_local_tzinfo()
    assert isinstance(tz, ZoneInfo)


def test_utc_str():
    dt = datetime(2020, 8, 3, 18, 0, 0)
    assert utc_str(dt) == "2020-08-03T18:00:00Z"
    assert utc_str(dt.replace(tzinfo=ZoneInfo("CET"))) == "2020-08-03T16:00:00Z"
    assert utc_str(dt.replace(tzinfo=ZoneInfo("GMT"))) == "2020-08-03T18:00:00Z"
    assert utc_str(dt.replace(tzinfo=ZoneInfo("US/Eastern"))) == "2020-08-03T22:00:00Z"
    assert utc_str(dt.replace(tzinfo=ZoneInfo("US/Pacific"))) == "2020-08-04T01:00:00Z"


def test_replace_env_vars():
    json = {
        "foo": {
            "foo": "$(BAR) $(BAR)",
            "bar": ["$(BAZ)"],
        },
        "bar-$(BAR)": ["$(BAR)", "$(BAZ)"],
    }
    env = {"BAR": "bar"}
    assert replace_env_vars(json, env) == {
        "foo": {
            "foo": "bar bar",
            "bar": ["$(BAZ)"],
        },
        "bar-$(BAR)": ["bar", "$(BAZ)"],
    }

    assert replace_env_vars(json, env, keep_unresolved=False) == {
        "foo": {
            "foo": "bar bar",
            "bar": [],
        },
        "bar-$(BAR)": ["bar"],
    }


def test_resolve_env_do_not_drop_nulls():
    a = {"a": "foo", "b": None}

    assert replace_env_vars(a, environment={}, keep_unresolved=False) == {"a": "foo", "b": None}
    assert replace_env_vars("$(FOO)", environment={}, keep_unresolved=True) == "$(FOO)"
    assert replace_env_vars("$(FOO)", environment={}, keep_unresolved=False) is None


def test_merge_json_elements():
    a = {"a": {"foo": {"first": "first", "last": "laaaast"}}, "b": {"bar": 123}, "c": [6, 7]}
    b = {"a": {"foo": {"last": "last"}}, "b": {"baz": 456}, "c": [8, 9]}

    assert merge_json_elements(a, b) == {
        "a": {"foo": {"first": "first", "last": "last"}},
        "b": {"bar": 123, "baz": 456},
        "c": [8, 9],
    }

    update = {
        "a": {
            "aa": 1,
            "ab": 2,
        },
        "b": {
            "ba": 3,
            "bb": {
                "bba": 4,
            },
        },
    }
    default = {"b": {"ba": 42}, "c": 2}

    def merge_fn(existing, update):
        if existing is None:
            return None
        return update

    assert merge_json_elements(default, update, merge_fn) == {"b": {"ba": 3}, "c": 2}


def test_do_not_drop_nulls():
    a = {"a": "foo", "b": None}

    b = {}
    assert merge_json_elements(a, b) == {"a": "foo", "b": None}

    c = {"a": "bar", "b": None}
    assert merge_json_elements(a, c) == {"a": "bar", "b": None}

    d = {"a": "bar", "b": "baz"}
    assert merge_json_elements(a, d) == {"a": "bar", "b": "baz"}


def test_drop_deleted_attributes():
    to_be_cleaned = {
        "foo": {
            "foo": "foo bar",
            "bar": ["bar"],
            "baz": "baz",
        },
        "bar": ["bar", "baz"],
        "baz": [1, 2, 3],
        "foobar": {
            "foo": "foo",
        },
    }

    reference = {
        "foo": {
            "foo": "1234",
            "bar": ["1234"],
        },
        "baz": [],
        "foobar": {
            "foo": "1234",
        },
    }

    assert drop_deleted_attributes(to_be_cleaned, reference) == {
        "foo": {
            "foo": "foo bar",
            "bar": ["bar"],
        },
        "baz": [1, 2, 3],
        "foobar": {
            "foo": "foo",
        },
    }


@contextmanager
def replace_stdin(input_data: str):
    with tempfile.TemporaryFile(mode="w+t") as temp_file:
        temp_file.write(input_data)
        temp_file.seek(0)
        original_stdin = sys.stdin
        sys.stdin = temp_file
        try:
            yield
        finally:
            sys.stdin = original_stdin


@pytest.mark.parametrize(
    "input_data, expected_output",
    [
        ("line1\nline2\nline3", ["line1", "line2", "line3"]),
        ("line1\r\nline2\r\nline3", ["line1", "line2", "line3"]),
        ("", []),
    ],
)
def test_sync_stdin_generator(input_data, expected_output):
    with replace_stdin(input_data):
        output = list(stdin_generator())
    assert output == expected_output


def test_freeze():
    tpl = (1, 2, 3)
    # hashable things are not touched
    assert id(tpl) == id(freeze(tpl))
    # dict to frozendict
    flat_dict = {"foo": "bar"}
    assert freeze(flat_dict) == frozendict(flat_dict)
    # nested too
    nested_dict = {"foo": flat_dict}
    assert freeze(nested_dict) == frozendict({"foo": frozendict(flat_dict)})
    # and a list to tupple
    assert freeze([1, 2]) == (1, 2)



from resotolib.baseplugin import BaseCollectorPlugin
from resotolib.args import ArgumentParser
from resotolib.baseresources import BaseResource
from typing import ClassVar, Any
from attrs import define


@define(eq=False, slots=False)
class SomeTestResource(BaseResource):
    kind: ClassVar[str] = "some_test_resource"

    def delete(self, graph) -> bool:
        return False


class SomeTestPlugin(BaseCollectorPlugin):
    cloud = "test"

    def collect(self, **kwargs: Any) -> None:
        account = SomeTestResource(id="Example Account", tags={})
        self.graph.add_resource(self.root, account)

    @staticmethod
    def add_args(arg_parser: ArgumentParser) -> None:
        arg_parser.add_argument(
            "--example-region",
            help="Example Region",
            dest="example_region",
            type=str,
            default=None,
            nargs="+",
        )


def test_plugin():
    plugin = SomeTestPlugin()
    plugin.collect()
    assert len(plugin.graph.nodes) == 2



import os
import time
from collections import defaultdict
from enum import Enum
from threading import Thread, Lock
from typing import Callable, Iterable, Any, Dict

from resotolib.lock import RWLock
from resotolib.logger import log
from resotolib.types import Json


class EventType(Enum):
    """Defines Event Types

    Event().data definitions for EventType:
    STARTUP: None
    SHUTDOWN: {'reason': 'reason for shutdown', 'emergency': True/False}
    START_COLLECT: None
    COLLECT_BEGIN: resotolib.graph.Graph
    COLLECT_FINISH: resotolib.graph.Graph
    CLEANUP_PLAN: resotolib.graph.Graph
    CLEANUP_BEGIN: resotolib.graph.Graph
    CLEANUP_FINISH: resotolib.graph.Graph
    PROCESS_BEGIN: resotolib.graph.Graph
    PROCESS_FINISH: resotolib.graph.Graph
    GENERATE_METRICS: resotolib.graph.Graph
    """

    STARTUP = "startup"
    SHUTDOWN = "shutdown"
    START_COLLECT = "start_collect"
    COLLECT_BEGIN = "collect_begin"
    COLLECT_FINISH = "collect_finish"
    CLEANUP_PLAN = "cleanup_plan"
    CLEANUP_BEGIN = "cleanup_begin"
    CLEANUP_FINISH = "cleanup_finish"
    PROCESS_BEGIN = "process_begin"
    PROCESS_FINISH = "process_finish"
    GENERATE_METRICS = "generate_metrics"


class Event:
    """An Event"""

    def __init__(self, event_type: EventType, data: Json) -> None:
        self.event_type = event_type
        self.data = data


_events: Dict[EventType, Dict[Callable[[Event], None], Any]] = defaultdict(dict)
_events_lock = RWLock()


def event_listener_registered(event_type: EventType, listener: Callable[[Event], None]) -> bool:
    """Return whether listener is registered to event"""
    if _events is None:
        return False
    return event_type in _events.keys() and listener in _events[event_type].keys()


def dispatch_event(event: Event, blocking: bool = False) -> None:
    """Dispatch an Event"""
    waiting_str = "" if blocking else "not "
    log.debug(f"Dispatching event {event.event_type.name} and {waiting_str}waiting for" " listeners to return")

    if event.event_type not in _events.keys():
        return

    with _events_lock.read_access:
        # Event listeners might unregister themselves during event dispatch
        # so we will work on a shallow copy while processing the current event.
        listeners = dict(_events[event.event_type])

    threads = {}
    for listener, listener_data in listeners.items():
        try:
            if listener_data["pid"] != os.getpid():
                continue

            if listener_data["one-shot"] and not listener_data["lock"].acquire(blocking=False):
                log.error(f"Not calling one-shot listener {listener} of type" f" {type(listener)} - can't acquire lock")
                continue

            log.debug(
                f"Calling listener {listener} of type {type(listener)}" f" (blocking: {listener_data['blocking']})"
            )
            thread_name = f"{event.event_type.name.lower()}_event" f"-{getattr(listener, '__name__', 'anonymous')}"
            t = Thread(target=listener, args=[event], name=thread_name)
            if blocking or listener_data["blocking"]:
                threads[t] = listener
            t.start()
        except Exception:
            log.exception("Caught unhandled event callback exception")
        finally:
            if listener_data["one-shot"]:
                log.debug(
                    f"One-shot specified for event {event.event_type.name} "
                    f"listener {listener} - removing event listener"
                )
                remove_event_listener(event.event_type, listener)
                listener_data["lock"].release()

    start_time = time.time()
    for thread, listener in threads.items():
        timeout = start_time + listeners[listener]["timeout"] - time.time()
        if timeout < 1:
            timeout = 1
        log.debug(f"Waiting up to {timeout:.2f}s for event listener {thread.name} to finish")
        thread.join(timeout)
        log.debug(f"Event listener {thread.name} finished (timeout: {thread.is_alive()})")


def add_event_listener(
    event_type: EventType,
    listener: Callable[[Event], None],
    blocking: bool = False,
    timeout: int = 900,
    one_shot: bool = False,
) -> bool:
    """Add an Event Listener"""
    if not callable(listener):
        log.error(f"Error registering {listener} of type {type(listener)} with event" f" {event_type.name}")
        return False

    log.debug(f"Registering {listener} with event {event_type.name}" f" (blocking: {blocking}, one-shot: {one_shot})")
    with _events_lock.write_access:
        if not event_listener_registered(event_type, listener):
            _events[event_type][listener] = {
                "blocking": blocking,
                "timeout": timeout,
                "one-shot": one_shot,
                "lock": Lock(),
                "pid": os.getpid(),
            }
            return True
        return False


def remove_event_listener(event_type: EventType, listener: Callable[[Event], None]) -> bool:
    """Remove an Event Listener"""
    with _events_lock.write_access:
        if event_listener_registered(event_type, listener):
            log.debug(f"Removing {listener} from event {event_type.name}")
            del _events[event_type][listener]
            if len(_events[event_type]) == 0:
                del _events[event_type]
            return True
        return False


def list_event_listeners() -> Iterable[str]:
    with _events_lock.read_access:
        for event_type, listeners in _events.items():
            for listener, listener_data in listeners.items():
                yield (
                    f"{event_type.name}: {listener}, "
                    f"blocking: {listener_data['blocking']}, "
                    f"one-shot: {listener_data['one-shot']}"
                )



import requests
import json
from typing import Dict, Tuple, List, Optional, cast, Union
from resotolib.core import resotocore
from resotolib.args import ArgumentParser
from resotolib.jwt import encode_jwt_to_headers
from resotolib.logger import log
from resotolib.types import Json


def default_args(
    resotocore_uri: Optional[str] = None, psk: Optional[str] = None, headers: Optional[Dict[str, str]] = None
) -> Tuple[str, Optional[str], Dict[str, str]]:
    if resotocore_uri is None:
        resotocore_uri = resotocore.http_uri
    if psk is None:
        psk = getattr(ArgumentParser.args, "psk", None)
    if headers is None:
        headers = {}
    if psk is not None:
        encode_jwt_to_headers(headers, {}, psk)
    return resotocore_uri, psk, headers


class ConfigNotFoundError(AttributeError):
    pass


def get_configs(
    resotocore_uri: Optional[str] = None, psk: Optional[str] = None, verify: Optional[str] = None
) -> List[Json]:
    resotocore_uri, psk, headers = default_args(resotocore_uri, psk)

    log.debug("Getting configs")
    r = requests.get(f"{resotocore_uri}/configs", headers=headers, verify=verify)
    if r.status_code == 200:
        return cast(List[Json], r.json())
    raise RuntimeError(f"Error getting configs: {r.content.decode('utf-8')}")


def get_config(
    config_id: str,
    resotocore_uri: Optional[str] = None,
    psk: Optional[str] = None,
    verify: Optional[str] = None,
) -> Tuple[Json, str]:
    resotocore_uri, psk, headers = default_args(resotocore_uri, psk)

    log.debug(f"Getting config {config_id}")

    params = {
        "separate_overrides": "true",  # we don not to have a single config with everything merged into it
        "apply_overrides": "true",  # apply the overrides to the config
        "resolve_env_vars": "true",  # and resolve any environment variables
        "include_raw_config": "true",  # also include the db version of the config
    }

    r = requests.get(
        f"{resotocore_uri}/config/{config_id}",
        headers=headers,
        verify=verify,
        params=params,
    )
    if r.status_code == 200:
        revision = r.headers.get("Resoto-Config-Revision", "unknown")
        return r.json(), revision
    elif r.status_code == 404:
        raise ConfigNotFoundError(f"Config {config_id} does not exist")
    raise RuntimeError(f"Error getting config {config_id}: {r.content.decode('utf-8')}")


def set_config(
    config_id: str,
    config: Json,
    resotocore_uri: Optional[str] = None,
    psk: Optional[str] = None,
    verify: Union[str, bool, None] = None,
) -> str:
    resotocore_uri, psk, headers = default_args(resotocore_uri, psk)

    log.debug(f"Storing config {config_id}")
    r = requests.put(
        f"{resotocore_uri}/config/{config_id}",
        json=config,
        headers=headers,
        verify=verify,
    )
    if r.status_code == 200:
        revision = r.headers.get("Resoto-Config-Revision", "unknown")
        return revision
    raise RuntimeError(f"Error storing config {config_id}: {r.content.decode('utf-8')}")


def delete_config(
    config_id: str,
    resotocore_uri: Optional[str] = None,
    psk: Optional[str] = None,
    verify: Union[str, bool, None] = None,
) -> bool:
    resotocore_uri, psk, headers = default_args(resotocore_uri, psk)

    log.debug(f"Deleting config {config_id}")
    r = requests.delete(f"{resotocore_uri}/config/{config_id}", headers=headers, verify=verify)
    if r.status_code == 204:
        return True
    raise RuntimeError(f"Error deleting config {config_id}: {r.content.decode('utf-8')}")


def update_config_model(
    model: List[Json],
    resotocore_uri: Optional[str] = None,
    psk: Optional[str] = None,
    verify: Union[str, bool, None] = None,
) -> bool:
    headers = {"Content-Type": "application/json"}
    resotocore_uri, psk, headers = default_args(resotocore_uri, psk, headers=headers)
    model_uri = f"{resotocore_uri}/configs/model"
    model_json = json.dumps(model, indent=4)

    log.debug("Updating config model")
    r = requests.patch(model_uri, data=model_json, headers=headers, verify=verify)
    if r.status_code != 200:
        log.error(r.content)
        raise RuntimeError(f"Failed to update model: {r.content.decode('utf-8')}")
    return True



import time
import requests
import warnings
from resotolib.logger import log
from resotolib.args import ArgumentParser
from urllib.parse import urlparse, ParseResult
from typing import Optional, Dict


class CLIEnvelope:
    """
    Envelope fields that are used by the CLI.
    Those fields are encoded as HTTP Headers into the HTTP response.
    """

    # Defines the action that should be performed.
    # Use cases:
    # - "edit": A file that is returned from the core should be opened in an editor.
    #           The result of the edit should be sent back to the core, identified by the "command" envelope field.
    action = "Resoto-Shell-Action"
    # Defines the command that should be executed after the edit was performed.
    command = "Resoto-Shell-Command"
    # Do not add this command to the shell history.
    no_history = "Resoto-Shell-No-History"


def add_args(arg_parser: ArgumentParser) -> None:
    arg_parser.add_argument(
        "--resotocore-uri",
        help="resotocore URI (default: https://localhost:8900)",
        default="https://localhost:8900",
        dest="resotocore_uri",
    )


def resotocore_is_up(resotocore_uri: str, timeout: int = 5, headers: Optional[Dict[str, str]] = None) -> bool:
    ready_uri = f"{resotocore_uri}/system/ready"
    try:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            response = requests.get(ready_uri, timeout=timeout, verify=False, headers=headers)
            if response.status_code == 200:
                return True
    except Exception:
        pass
    return False


def wait_for_resotocore(resotocore_uri: str, timeout: int = 300, headers: Optional[Dict[str, str]] = None) -> None:
    start_time = time.time()
    core_up = False
    wait_time: float = -1
    remaining_wait: float = timeout
    waitlog = log.info
    while wait_time < timeout:
        if resotocore_is_up(resotocore_uri, headers=headers):
            core_up = True
            break
        else:
            waitlog(f"Waiting up to {remaining_wait:.2f}s for resotocore" f" to come online at {resotocore_uri}")
            waitlog = log.debug
        time.sleep(2)
        wait_time = time.time() - start_time
        remaining_wait = timeout - wait_time
    if not core_up:
        raise TimeoutError(f"resotocore not ready after {wait_time:.2f} seconds")


class ResotocoreURI:
    def __init__(self, resotocore_uri: Optional[str] = None) -> None:
        self.resotocore_uri = resotocore_uri

    @property
    def uri(self) -> ParseResult:
        if self.resotocore_uri is None:
            resotocore_uri = getattr(ArgumentParser.args, "resotocore_uri", None)
            if resotocore_uri is not None:
                resotocore_uri = str(resotocore_uri).rstrip("/")
            else:
                resotocore_uri = "https://localhost:8900"
        else:
            resotocore_uri = self.resotocore_uri
        if resotocore_uri is None:
            raise AttributeError("resotocore_uri is not set")
        return urlparse(resotocore_uri)  # type: ignore

    @property
    def http_uri(self) -> str:
        return f"{self.uri.scheme}://{self.uri.netloc}{self.uri.path}"

    @property
    def ws_uri(self) -> str:
        scheme = "ws"
        if self.uri.scheme == "https":
            scheme = "wss"
        return f"{scheme}://{self.uri.netloc}{self.uri.path}"

    @property
    def is_secure(self) -> bool:
        return self.uri.scheme == "https"


resotocore = ResotocoreURI()



from typing import List, Any, Generator, Hashable

from networkx import MultiDiGraph, DiGraph, connected_components


def dependent_node_iterator(
    in_graph: DiGraph,
) -> Generator[List[Any], None, None]:
    """
    Create a generator that produces a list of nodes.
    Each generation of nodes only depend on previously generated nodes.
    """

    def successor_it(g: DiGraph) -> Generator[List[Any], None, None]:
        visited = set()
        # start with all roots of the sub-graph
        to_emit = {n for n, d in g.in_degree if d == 0}

        # make sure a node is only selected if it is not visited already
        # and all predecessors have been visited already
        def allowed(nid: Hashable) -> bool:
            pred = g.predecessors(nid)
            req = [n for n in pred if n != nid and n not in visited]
            return nid not in visited and not req

        while to_emit:
            # emit the related node
            yield [nid for nid in to_emit]
            # add all nodes as visited
            visited.update(to_emit)
            # get all successors
            to_emit = {succ for nid in to_emit for succ in g.successors(nid) if allowed(succ)}

    if isinstance(in_graph, MultiDiGraph):
        raise RuntimeError("MultiDiGraph not supported")
    # reverse the directed graph -> a leaf becomes a root
    graph = in_graph.reverse()
    # find all islands and create a generator per island
    generators = [
        successor_it(graph.subgraph(island_nodes))
        for island_nodes in connected_components(graph.to_undirected(as_view=True))
    ]
    # concatenate the result of the generators
    while generators:
        nxt = []
        for e in generators:
            try:
                nxt.extend(next(e))
            except StopIteration:
                generators.remove(e)
        if nxt:
            yield nxt



import gc
import os
import re
import time
import sys
from typing import Optional, Dict, List

import psutil
import threading
import subprocess
from resotolib.logger import log
from resotolib.event import dispatch_event, Event, EventType
from signal import signal, Signals, SIGTERM, SIGINT

from resotolib.utils import iec_size_format

try:
    import resource
    import fcntl
except ImportError:
    pass

try:
    from psutil import cpu_count
except ImportError:
    from os import cpu_count


parent_pid: Optional[int] = None
initial_dir: str = os.getcwd()


def restart() -> None:
    python_args = []
    if not getattr(sys, "frozen", False):
        python_args = subprocess._args_from_interpreter_flags()
    args = python_args + sys.argv

    path_prefix = "." + os.pathsep
    python_path = os.environ.get("PYTHONPATH", "")
    if sys.path[0] == "" and not python_path.startswith(path_prefix):
        os.environ["PYTHONPATH"] = path_prefix + python_path

    try:
        close_fds()
    except Exception:
        log.exception("Failed to FD_CLOEXEC all file descriptors")

    kill_children(SIGTERM, ensure_death=True)

    os.chdir(initial_dir)
    os.execv(sys.executable, [sys.executable] + args)
    log.fatal("Failed to restart - exiting")
    os._exit(1)


def delayed_exit(delay: int = 3) -> None:
    time.sleep(delay)
    os._exit(0)


def close_fds(safety_margin: int = 1024) -> None:
    """Set FD_CLOEXEC on all file descriptors except stdin, stdout, stderr

    Since there is a race between determining the max number of fds to close
    and actually closing them we are adding a safety margin.
    """
    if sys.platform == "win32":
        return

    open_fds = [f.fd for f in psutil.Process().open_files()]
    if len(open_fds) == 0:
        return

    num_open = max(open_fds)

    try:
        sc_open_max = os.sysconf("SC_OPEN_MAX")
    except AttributeError:
        sc_open_max = 1024

    num_close = min(num_open + safety_margin, sc_open_max)

    for fd in range(3, num_close):
        fd_cloexec(fd)


def fd_cloexec(fd: int) -> None:
    try:
        flags = fcntl.fcntl(fd, fcntl.F_GETFD)
    except IOError:
        return
    fcntl.fcntl(fd, fcntl.F_SETFD, flags | fcntl.FD_CLOEXEC)


def handler(sig, frame) -> None:
    """Handles Ctrl+c by letting the Collector() know to shut down"""
    current_pid = os.getpid()
    if current_pid == parent_pid:
        reason = f"Received shutdown signal {sig}"
        log.debug(f"Parent caught signal {sig} - dispatching shutdown event")
        # Dispatch shutdown event in parent process which also causes SIGTERM to be sent
        # to the process group and in turn causes the shutdown event in all child
        # processes.
        dispatch_event(Event(EventType.SHUTDOWN, {"reason": reason, "emergency": False}))
    else:
        reason = f"Received shutdown signal {sig} from parent process"
        log.debug(
            f"Child with PID {current_pid} shutting down" " - you might see exceptions from interrupted worker threads"
        )
        # Child's threads have 3s to shut down before the following thread will
        # shut them down hard.
        kt = threading.Thread(target=delayed_exit, name="shutdown")
        kt.start()
        # Dispatch shutdown event in child process
        dispatch_event(
            Event(EventType.SHUTDOWN, {"reason": reason, "emergency": False}),
            blocking=False,
        )
        sys.exit(0)


def initializer() -> None:
    signal(SIGINT, handler)
    signal(SIGTERM, handler)
    increase_limits()


def collector_initializer(nice_level: int = 5) -> None:
    initializer()
    try:
        os.nice(nice_level)
    except Exception:
        pass


def set_thread_name(thread_name: str = "resoto") -> None:
    threading.current_thread().name = thread_name


def emergency_shutdown(reason: str = "") -> None:
    log.fatal(f"EMERGENCY SHUTDOWN: {reason}")
    for p in psutil.Process().children(recursive=True):
        p.kill()
    psutil.Process().kill()


def kill_children(
    signal: Signals = SIGTERM, ensure_death: bool = False, timeout: int = 3, process_pid: Optional[int] = None
) -> None:
    procs = psutil.Process(process_pid).children(recursive=True)
    num_children = len(procs)
    if num_children == 0:
        return
    elif num_children == 1:
        log_suffix = ""
    else:
        log_suffix = "ren"

    log.debug(f"Sending {signal.name} to {num_children} child{log_suffix}.")
    for p in procs:
        try:
            if signal == SIGTERM:
                p.terminate()
            else:
                p.send_signal(signal)
        except psutil.NoSuchProcess:
            pass

    if ensure_death:
        _, alive = psutil.wait_procs(procs, timeout=timeout)
        for p in alive:
            log.debug(f"Child with PID {p.pid} is still alive, sending SIGKILL")
            try:
                p.kill()
            except psutil.NoSuchProcess:
                pass


def increase_limits() -> None:
    if sys.platform != "linux":
        return
    for limit_name in ("RLIMIT_NOFILE", "RLIMIT_NPROC"):
        soft_limit, hard_limit = resource.getrlimit(getattr(resource, limit_name))
        log.debug(f"Current {limit_name} soft: {soft_limit} hard: {hard_limit}")
        try:
            if soft_limit < hard_limit:
                log.debug(f"Increasing {limit_name} {soft_limit} -> {hard_limit}")
                resource.setrlimit(getattr(resource, limit_name), (hard_limit, hard_limit))
        except ValueError:
            log.error(f"Failed to increase {limit_name} {soft_limit} -> {hard_limit}")


def num_default_threads(num_min_threads: int = 2) -> int:
    count = num_min_threads
    try:
        # try to get the number of usable cores first
        count = len(os.sched_getaffinity(0))
    except AttributeError:
        try:
            count = cpu_count()
        except Exception:
            pass
    if not isinstance(count, int):
        count = num_min_threads
    return max(count, num_min_threads)


def get_process_info(pid: int = None, proc: str = "/proc") -> Dict:
    if sys.platform != "linux":
        return {}
    if pid is None:
        pid = os.getpid()
    process_info = {}
    try:
        with open(os.path.join(proc, str(pid), "status"), "r") as status:
            for line in status:
                k, v = line.split(":", 1)
                v = re.sub("[ \t]+", " ", v.strip())
                process_info[k.lower()] = v
        for limit_name in ("NOFILE", "NPROC"):
            process_info[f"RLIMIT_{limit_name}".lower()] = resource.getrlimit(getattr(resource, f"RLIMIT_{limit_name}"))
    except (PermissionError, FileNotFoundError):
        pass
    return process_info


def get_file_descriptor_info(pid: int = None, proc: str = "/proc") -> Dict:
    if sys.platform != "linux":
        return {}
    if pid is None:
        pid = os.getpid()
    pid = str(pid)
    file_descriptor_info = {}
    try:
        for entry in os.listdir(os.path.join(proc, pid, "fd")):
            entry_path = os.path.join(proc, pid, "fd", entry)
            if os.path.islink(entry_path):
                file_descriptor_info[entry] = {}
                target = os.readlink(entry_path)
                file_descriptor_info[entry]["target"] = target
    except (PermissionError, FileNotFoundError):
        pass
    return file_descriptor_info


def get_pid_list(proc: str = "/proc") -> List:
    if sys.platform != "linux":
        return []
    pids = []
    for entry in os.listdir(proc):
        try:
            if os.path.isdir(os.path.join(proc, entry)) and entry.isdigit():
                pids.append(int(entry))
        except (PermissionError, FileNotFoundError):
            pass
    return pids


def get_all_process_info(pid: int = None, proc: str = "/proc") -> Dict:
    if sys.platform != "linux":
        return {}
    if pid is None:
        pid = os.getpid()
    process_info = {}
    process_info["parent"] = get_process_info(pid)
    process_info["parent"]["file_descriptors"] = get_file_descriptor_info(pid, proc)
    process_info["parent"]["num_file_descriptors"] = len(process_info["parent"]["file_descriptors"])
    process_info["children"] = get_child_process_info(pid, proc)
    for pid in process_info["children"]:
        process_info["children"][pid]["file_descriptors"] = get_file_descriptor_info(pid, proc)
        process_info["children"][pid]["num_file_descriptors"] = len(process_info["children"][pid]["file_descriptors"])
    return process_info


def get_child_process_info(pid: int = None, proc: str = "/proc") -> Dict:
    if sys.platform != "linux":
        return {}
    if pid is None:
        pid = os.getpid()
    child_process_info = {}
    for pid in get_pid_list(proc):
        process_info = get_process_info(pid)
        if process_info.get("ppid") == str(pid):
            child_process_info[pid] = dict(process_info)
    return child_process_info


def get_stats() -> Dict:
    try:
        stats = {
            "active_threads": threading.active_count(),
            "thread_names": [thread.name for thread in threading.enumerate()],
            "garbage_collector": gc.get_stats(),
            "process": get_all_process_info(),
        }
        if sys.platform == "linux":
            stats.update(
                {
                    "maxrss_parent_bytes": resource.getrusage(resource.RUSAGE_SELF).ru_maxrss * 1024,
                    "maxrss_children_bytes": resource.getrusage(resource.RUSAGE_CHILDREN).ru_maxrss * 1024,
                }
            )
        else:
            stats.update({"maxrss_parent_bytes": 0, "maxrss_children_bytes": 0})
        stats["maxrss_total_bytes"] = stats["maxrss_parent_bytes"] + stats["maxrss_children_bytes"]
        num_fds_parent = stats["process"].get("parent", {}).get("num_file_descriptors", 0)
        num_fds_children = sum([v["num_file_descriptors"] for v in stats["process"].get("children", {}).values()])
        stats.update(
            {
                "maxrss_parent_human_readable": iec_size_format(stats["maxrss_parent_bytes"]),
                "maxrss_children_human_readable": iec_size_format(stats["maxrss_children_bytes"]),
                "maxrss_total_human_readable": iec_size_format(stats["maxrss_total_bytes"]),
                "num_fds_parent": num_fds_parent,
                "num_fds_children": num_fds_children,
                "num_fds_total": num_fds_parent + num_fds_children,
            }
        )
    except Exception:
        log.exception("Error while trying to get stats")
        return {}
    else:
        return stats


def log_stats(graph=None, garbage_collector_stats: bool = False) -> None:
    stats = get_stats()
    try:
        log.debug(
            f"Stats: max rss parent: {stats['maxrss_parent_human_readable']},"
            f" children: {stats['maxrss_children_human_readable']},"
            f" fds: {stats['num_fds_total']}/"
            f"{stats['process'].get('parent', {}).get('rlimit_nofile', [0])[0]}"
            f" active threads {stats['active_threads']}:"
            f" {', '.join([thread for thread in stats['thread_names']])}"
        )
        if graph:
            log.debug(f"Graph Stats: {stats['graph_size_human_readable']}")
        if garbage_collector_stats:
            gc_stats = " | ".join(
                [
                    (
                        f"Gen {i}: collections {data.get('collections')}, "
                        f"collected {data.get('collected')}, "
                        f"uncollectable {data.get('uncollectable')}"
                    )
                    for i, data in enumerate(stats["garbage_collector"])
                ]
            )
            log.debug(f"Garbage Collector Stats: {gc_stats}")
    except Exception:
        log.exception("Error while trying to log stats")



from __future__ import annotations

from abc import abstractmethod, ABC
from typing import List, Optional, Any, Dict, Callable

from attr import define, field, evolve
from resotolib.tree import Tree, Node

from resotolib.types import Json, JsonElement

_TreeRoot = "root"


@define
class ProgressInfo:
    current: int
    total: int

    @property
    def percentage(self) -> int:
        return int(self.current * 100 / self.total) if self.total > 0 else 0

    @property
    def done(self) -> bool:
        return self.current == self.total


@define
class Progress(ABC):
    name: str
    path: List[str] = field(kw_only=True, factory=list)

    @abstractmethod
    def overall_progress(self) -> ProgressInfo:
        pass

    @abstractmethod
    def update_progress(self, progress: Progress) -> Progress:
        pass

    @abstractmethod
    def mark_done(self) -> Progress:
        pass

    @property
    def percentage(self) -> int:
        return self.overall_progress().percentage

    @staticmethod
    def from_progresses(name: str, progresses: List[Progress]) -> ProgressTree:
        updated = ProgressTree(name)
        for p in progresses:
            updated.add_progress(p)
        return updated

    def to_json(self, key: Optional[Callable[[Progress], Any]] = None) -> Json:
        p = {"path": self.path} if self.path else {}
        if isinstance(self, ProgressDone):
            return {
                "kind": "progress",
                "name": self.name,
                **p,
                "current": self.current,
                "total": self.total,
            }
        elif isinstance(self, ProgressTree):
            node_iter = (part.data for part in self.sub_tree.all_nodes_itr() if part.data is not None)
            nodes: List[Progress] = sorted(node_iter, key=key) if key else list(node_iter)
            return {"kind": "tree", "name": self.name, **p, "parts": [part.to_json() for part in nodes]}
        else:
            raise AttributeError("No handler to marshal progress")

    def info_json(self) -> JsonElement:
        if isinstance(self, ProgressDone):
            if self.total == self.current:
                return "done"
            elif self.current == 0:
                return "in progress"
            else:
                return f"{self.percentage}%"
        elif isinstance(self, ProgressTree):
            cloned = Tree(self.sub_tree.subtree(self.sub_tree.root), deep=True)

            def level_info(nd: Node) -> Json:
                if dt := nd.data:
                    return dt.info_json()  # type: ignore
                else:
                    return {child.tag: level_info(child) for child in cloned.children(nd.identifier)}

            return {self.name: level_info(cloned.get_node(cloned.root))}
        else:
            raise AttributeError("No handler to marshal progress")

    @staticmethod
    def from_json(json: Json) -> Progress:
        name = json["name"]
        path = json.get("path", [])
        if json["kind"] == "progress":
            return ProgressDone(name, json["current"], json["total"], path=path)
        elif json["kind"] == "tree":
            tree = ProgressTree(name, path=path)
            for part in json["parts"]:
                tree.add_progress(Progress.from_json(part))
            return tree
        else:
            raise AttributeError("No handler to unmarshal progress")


@define
class ProgressDone(Progress):
    current: int
    total: int

    def __attrs_post_init__(self) -> None:
        if self.total <= 0:
            raise ValueError("total must be greater than 0")
        if self.current > self.total:
            raise ValueError(f"current ({self.current}) > total ({self.total})")

    def __str__(self) -> str:
        return f"{self.current}/{self.total}"

    def overall_progress(self) -> ProgressInfo:
        return ProgressInfo(self.current, self.total)

    def update_progress(self, progress: Progress) -> Progress:
        return Progress.from_progresses(_TreeRoot, [self, progress])

    def mark_done(self) -> Progress:
        return evolve(self, current=self.total)


@define(eq=False)
class ProgressTree(Progress):
    sub_tree: Tree = field(factory=Tree)

    def __attrs_post_init__(self) -> None:
        if not self.sub_tree.root:
            self.sub_tree.create_node(_TreeRoot, _TreeRoot)

    def __eq__(self, other: Any) -> bool:
        def data_nodes(tree: Tree) -> Dict[str, Progress]:
            return {nid: node.data for nid, node in tree.nodes.items() if node.data is not None}

        if isinstance(other, ProgressTree):
            return data_nodes(self.sub_tree) == data_nodes(other.sub_tree)
        else:
            return False

    def sub_progress(self, nid: str) -> Optional[Progress]:
        nid = nid if nid.startswith(_TreeRoot) else _TreeRoot + "." + nid
        node = self.sub_tree.get_node(nid)
        if node is None:
            return None
        elif node.data is not None:
            return node.data  # type: ignore
        else:
            return evolve(self, sub_tree=Tree(self.sub_tree.subtree(nid), deep=True))

    def has_path(self, nid: str) -> bool:
        nid = nid if nid.startswith(_TreeRoot) else _TreeRoot + "." + nid
        return nid in self.sub_tree

    def by_path(self, nid: str) -> Optional[Progress]:
        nid = nid if nid.startswith(_TreeRoot) else _TreeRoot + "." + nid
        node = self.sub_tree.get_node(nid)
        return node.data if node and node.data else None

    def overall_progress(self) -> ProgressInfo:
        def sub_progress_info(nid: str) -> ProgressInfo:
            node = self.sub_tree.get_node(nid)
            # either the node is a data node or a tree node with children
            if node.data is not None:
                return node.data.overall_progress()  # type: ignore
            else:
                parts = [sub_progress_info(child.identifier) for child in self.sub_tree.children(nid)]
                total_max = max(parts, key=lambda x: x.total).total if parts else 1
                current = 0
                total = 0
                for info in parts:
                    current += int(info.current * total_max / max(1, info.total))
                    total += total_max
                return ProgressInfo(current, total)

        return sub_progress_info(self.sub_tree.root)  # type: ignore

    def mark_done(self) -> Progress:
        cloned = evolve(self, sub_tree=Tree(self.sub_tree.subtree(self.sub_tree.root), deep=True))
        for node in cloned.sub_tree.all_nodes():
            if node.data is not None:
                node.data = node.data.mark_done()
        return cloned

    def update_progress(self, progress: Progress) -> Progress:
        cloned = evolve(self, sub_tree=Tree(self.sub_tree.subtree(self.sub_tree.root), deep=True))
        cloned.add_progress(progress)
        return cloned

    def add_progress(self, progress: Progress) -> None:
        last: str = self.sub_tree.root  # type: ignore
        last_path = last
        path = last
        for part in progress.path:
            path += "." + part
            if path not in self.sub_tree:  # if the path does not exist, create it
                self.sub_tree.create_node(part, path, parent=last_path)
            elif self.sub_tree[path].data is not None:  # if the path contains a value: remove it
                self.sub_tree[path].data = None
            last_path = path

        nid = path + "." + progress.name
        if nid in self.sub_tree and nid != self.sub_tree.root:
            self.sub_tree.remove_node(nid)
        if isinstance(progress, ProgressDone):
            self.sub_tree.create_node(progress.name, nid, parent=last_path, data=progress)
        elif isinstance(progress, ProgressTree):
            rp = nid.split(".")[1:]  # the relative path of the progress tree inside current tree (without root)
            for node in progress.sub_tree.leaves():
                if data := node.data:
                    np = node.identifier.split(".")[1:-1]  # relative path of leaf: no root not current node
                    self.add_progress(evolve(data, path=rp + np + data.path))

    def copy(self) -> ProgressTree:
        return evolve(self, sub_tree=Tree(self.sub_tree.subtree(self.sub_tree.root), deep=True))



import asyncio
import logging
import sys
from ssl import SSLContext
from typing import Optional, Union, Awaitable, Callable, Type, List, cast

from aiohttp.abc import AbstractAccessLogger
from aiohttp.log import access_logger

# noinspection PyProtectedMember
from aiohttp.web import HostSequence, _cancel_tasks
from aiohttp.web_app import Application
from aiohttp.web_log import AccessLogger
from aiohttp.web_runner import GracefulExit, BaseSite, TCPSite, AppRunner


# This method is derived from aiohttp.web.run_app with additional steps:
# - it allows a callable that is executed on shutdown.
# - it does not swallow terminal exceptions
# - it exposes a http and https port
def run_app(
    app: Union[Application, Awaitable[Application]],
    on_shutdown: Callable[[], Awaitable[None]],
    host: Union[str, HostSequence],
    https_port: Optional[int],
    http_port: Optional[int],
    default_port: int,
    *,
    shutdown_timeout: float = 60.0,
    keepalive_timeout: float = 75.0,
    ssl_context: Optional[SSLContext] = None,
    print_cmd: Callable[..., None] = print,
    backlog: int = 128,
    access_log_class: Type[AbstractAccessLogger] = AccessLogger,
    access_log_format: str = AccessLogger.LOG_FORMAT,
    access_log: Optional[logging.Logger] = access_logger,
    handle_signals: bool = True,
    reuse_address: Optional[bool] = None,
    reuse_port: Optional[bool] = None,
    loop: Optional[asyncio.AbstractEventLoop] = None,
) -> None:
    """Run an app locally"""
    if loop is None:
        loop = asyncio.new_event_loop()

    # Configure if and only if in debugging mode and using the default logger
    if loop.get_debug() and access_log and access_log.name == "aiohttp.access":
        if access_log.level == logging.NOTSET:
            access_log.setLevel(logging.DEBUG)
        if not access_log.hasHandlers():
            access_log.addHandler(logging.StreamHandler())

    main_task = loop.create_task(
        _run_app(
            app,
            host=host,
            https_port=https_port,
            http_port=http_port,
            default_port=default_port,
            shutdown_timeout=shutdown_timeout,
            keepalive_timeout=keepalive_timeout,
            ssl_context=ssl_context,
            print=print_cmd,
            backlog=backlog,
            access_log_class=access_log_class,
            access_log_format=access_log_format,
            access_log=access_log,
            handle_signals=handle_signals,
            reuse_address=reuse_address,
            reuse_port=reuse_port,
        )
    )

    try:
        asyncio.set_event_loop(loop)
        loop.run_until_complete(main_task)
    except (GracefulExit, KeyboardInterrupt) as ex:  # pragma: no cover
        raise ex
    finally:
        loop.run_until_complete(on_shutdown())
        _cancel_tasks({main_task}, loop)
        _cancel_tasks(asyncio.all_tasks(loop), loop)
        loop.run_until_complete(loop.shutdown_asyncgens())
        loop.close()


async def _run_app(
    app: Union[Application, Awaitable[Application]],
    host: Union[str, HostSequence],
    https_port: Optional[int],
    http_port: Optional[int],
    default_port: int,
    *,
    shutdown_timeout: float = 60.0,
    keepalive_timeout: float = 75.0,
    ssl_context: Optional[SSLContext] = None,
    print: Optional[Callable[..., None]] = print,
    backlog: int = 128,
    access_log_class: Type[AbstractAccessLogger] = AccessLogger,
    access_log_format: str = AccessLogger.LOG_FORMAT,
    access_log: Optional[logging.Logger] = access_logger,
    handle_signals: bool = True,
    reuse_address: Optional[bool] = None,
    reuse_port: Optional[bool] = None,
) -> None:
    # An internal function to actually do all dirty job for application running
    if asyncio.iscoroutine(app):
        app = await app

    app = cast(Application, app)

    runner = AppRunner(
        app,
        handle_signals=handle_signals,
        access_log_class=access_log_class,
        access_log_format=access_log_format,
        access_log=access_log,
        keepalive_timeout=keepalive_timeout,
    )

    await runner.setup()

    sites: List[BaseSite] = []

    def with_port(port: int, ssl: Optional[SSLContext] = None) -> None:
        if isinstance(host, (str, bytes, bytearray, memoryview)):
            sites.append(
                TCPSite(
                    runner,
                    host,
                    port,
                    shutdown_timeout=shutdown_timeout,
                    ssl_context=ssl,
                    backlog=backlog,
                    reuse_address=reuse_address,
                    reuse_port=reuse_port,
                )
            )
        else:
            for h in host:
                sites.append(
                    TCPSite(
                        runner,
                        h,
                        port,
                        shutdown_timeout=shutdown_timeout,
                        ssl_context=ssl,
                        backlog=backlog,
                        reuse_address=reuse_address,
                        reuse_port=reuse_port,
                    )
                )

    try:
        if https_port is not None:
            with_port(https_port, ssl_context)
        if http_port is not None and http_port != https_port:
            with_port(http_port)
        if http_port is None and https_port is None:
            with_port(default_port)

        for site in sites:
            await site.start()

        if print:
            names = sorted(str(s.name) for s in runner.sites)
            print("======== Running on {} ========\n" "(Press CTRL+C to quit)".format(", ".join(names)))

        # sleep forever by 1 hour intervals,
        # on Windows before Python 3.8 wake up every 1 second to handle
        # Ctrl+C smoothly
        if sys.platform == "win32" and sys.version_info < (3, 8):
            delay = 1
        else:
            delay = 3600

        while True:
            await asyncio.sleep(delay)
    finally:
        await runner.cleanup()



import os
from typing import List

from resotolib.args import get_arg_parser, ArgumentParser, convert, NoneType
from resotolib.logger import add_args as logging_add_args
from resotolib.jwt import add_args as jwt_add_args


def test_args():
    assert ArgumentParser.args.does_not_exist is None

    os.environ["RESOTO_PSK"] = "changeme"
    arg_parser = get_arg_parser()
    logging_add_args(arg_parser)
    jwt_add_args(arg_parser)
    arg_parser.parse_args()
    assert ArgumentParser.args.verbose is False
    assert ArgumentParser.args.psk == "changeme"

    os.environ["RESOTO_PSK"] = ""
    os.environ["RESOTO_VERBOSE"] = "true"
    os.environ["RESOTO_TEST_INT"] = "123"
    os.environ["RESOTO_TEST_LIST0"] = "foobar"
    arg_parser = get_arg_parser()
    logging_add_args(arg_parser)
    jwt_add_args(arg_parser)
    arg_parser.add_argument(
        "--test-int",
        dest="test_int",
        type=int,
        default=0,
    )
    arg_parser.add_argument(
        "--test-list",
        dest="test_list",
        type=str,
        default=[],
        nargs="+",
    )

    arg_parser.parse_args()
    assert ArgumentParser.args.verbose is True
    assert ArgumentParser.args.psk is None
    assert ArgumentParser.args.test_int == 123
    assert ArgumentParser.args.test_list[0] == "foobar"


def test_convert() -> None:
    def make_a_list(s: str) -> List[str]:
        return s.split(",")

    # coercing works
    assert convert(None, NoneType) is None
    assert convert("3", int) == 3
    assert convert("3.4", float) == 3.4
    assert convert("true", bool) is True
    assert convert("false", bool) is False
    assert convert("123", complex) == complex(123)

    # coercing is not possible
    assert convert("no_int", int) == "no_int"
    assert convert("no_float", float) == "no_float"
    assert convert("no_complex", complex) == "no_complex"

    # does not know how to handle
    assert convert("args", ArgumentParser) == "args"

    # call a function
    assert convert("1,2,3,4", make_a_list) == ["1", "2", "3", "4"]



import logging
from resotolib.logger import log, JsonFormatter


def test_logging():
    assert type(log) is logging.Logger


def test_json_logging() -> None:
    format = JsonFormatter({"level": "levelname", "message": "message"})
    record = logging.getLogger().makeRecord("test", logging.INFO, "test", 1, "test message", (), None)
    assert format.format(record) == '{"level": "INFO", "message": "test message"}'



import json
from attrs import define, field
from typing import ClassVar, Optional, Dict, Mapping, Any
import sys
import os
import logging
from logging import (
    basicConfig,
    getLogger,
    setLoggerClass,
    DEBUG,
    INFO,
    WARNING,
    ERROR,
    CRITICAL,
    StreamHandler,
    Formatter,
    LogRecord,
    Logger,
)


from resotolib.args import ArgumentParser
from resotolib.types import Json

DEBUG2 = DEBUG - 1
DEBUG3 = DEBUG - 2
DEBUG4 = DEBUG - 3
DEBUG5 = DEBUG - 4
TRACE = DEBUG - 5

getLogger().setLevel(ERROR)
getLogger("resoto").setLevel(INFO)


def add_args(arg_parser: ArgumentParser) -> None:
    group = arg_parser.add_mutually_exclusive_group()
    group.add_argument(
        "--verbose",
        "-v",
        help="Verbose logging",
        dest="verbose",
        action="store_true",
        default=False,
    )
    group.add_argument(
        "--trace",
        help="Trage logging",
        dest="trace",
        action="store_true",
        default=False,
    )
    group.add_argument(
        "--quiet",
        help="Only log errors",
        dest="quiet",
        action="store_true",
        default=False,
    )


@define
class LoggingConfig:
    kind: ClassVar[str] = "logging"
    verbose: Optional[bool] = field(default=False, metadata={"description": "Verbose logging"})
    quiet: Optional[bool] = field(default=False, metadata={"description": "Only log errors"})


class JsonFormatter(Formatter):
    """
    Simple json log formatter.
    Inspired by: https://stackoverflow.com/questions/50144628/python-logging-into-file-as-a-dictionary-or-json
    """

    def __init__(
        self,
        fmt_dict: Mapping[str, str],
        time_format: str = "%Y-%m-%dT%H:%M:%S",
        static_values: Optional[Dict[str, str]] = None,
    ) -> None:
        super().__init__()
        self.fmt_dict = fmt_dict
        self.time_format = time_format
        self.static_values = static_values or {}
        self.__use_time = "asctime" in self.fmt_dict.values()

    def usesTime(self) -> bool:  # noqa: N802
        return self.__use_time

    def formatMessage(self, record: LogRecord) -> dict:  # type: ignore # noqa: N802
        return {fmt_key: record.__dict__[fmt_val] for fmt_key, fmt_val in self.fmt_dict.items()}

    def formatJsonMessage(self, record) -> Json:  # type: ignore # noqa: N802
        record.message = record.getMessage()

        if self.__use_time:
            record.asctime = self.formatTime(record, self.time_format)

        message_dict = self.formatMessage(record)
        message_dict.update(self.static_values)

        if record.exc_info:
            if not record.exc_text:
                record.exc_text = self.formatException(record.exc_info)

        if record.exc_text:
            message_dict["exception"] = record.exc_text

        if record.stack_info:
            message_dict["stack_info"] = self.formatStack(record.stack_info)
        return message_dict

    def format(self, record: LogRecord) -> str:
        message_dict = self.formatJsonMessage(record)
        return json.dumps(message_dict, default=str)


def setup_logger(
    proc: str,
    *,
    force: bool = True,
    verbose: bool = False,
    quiet: bool = False,
    level: Optional[str] = None,
    json_format: bool = True,
) -> None:
    # override log output via env var
    plain_text = os.environ.get("RESOTO_LOG_TEXT", "false").lower() == "true"
    if json_format and not plain_text:
        handler = StreamHandler()
        formatter = JsonFormatter(
            {
                "timestamp": "asctime",
                "level": "levelname",
                "message": "message",
                "pid": "process",
                "thread": "threadName",
            },
            static_values={"process": proc},
        )
        handler.setFormatter(formatter)
        basicConfig(handlers=[handler], force=force, level=level)
    else:
        log_format = f"%(asctime)s|{proc}|%(levelname)5s|%(process)d|%(threadName)10s  %(message)s"
        # allow to define the log format via env var
        log_format = os.environ.get("RESOTO_LOG_FORMAT", log_format)
        basicConfig(format=log_format, datefmt="%y-%m-%d %H:%M:%S", force=force)
    argv = sys.argv[1:]
    if level:
        getLogger("resoto").setLevel(level)
    elif "--trace" in argv or os.environ.get("RESOTO_TRACE", "false").lower() == "true":
        getLogger("resoto").setLevel(TRACE)
    elif verbose or "-v" in argv or "--verbose" in argv or os.environ.get("RESOTO_VERBOSE", "false").lower() == "true":
        getLogger("resoto").setLevel(DEBUG)
    elif quiet or "--quiet" in argv or os.environ.get("RESOTO_QUIET", "false").lower() == "true":
        getLogger().setLevel(WARNING)
        getLogger("resoto").setLevel(CRITICAL)


# via https://stackoverflow.com/a/35804945/92184
def add_logging_level(level_name: str, level_num: int, method_name: Optional[str] = None) -> None:
    """
    Comprehensively adds a new logging level to the `logging` module and the
    currently configured logging class.

    `levelName` becomes an attribute of the `logging` module with the value
    `levelNum`. `methodName` becomes a convenience method for both `logging`
    itself and the class returned by `logging.getLoggerClass()` (usually just
    `logging.Logger`). If `methodName` is not specified, `levelName.lower()` is
    used.

    To avoid accidental clobberings of existing attributes, this method will
    raise an `AttributeError` if the level name is already an attribute of the
    `logging` module or if the method name is already present

    Example
    -------
    >>> add_logging_level("TRACE", logging.DEBUG - 5)
    >>> logging.getLogger(__name__).setLevel("TRACE")
    >>> logging.getLogger(__name__).trace("that worked")
    >>> logging.trace("so did this")
    >>> logging.TRACE
    5

    """
    if not method_name:
        method_name = level_name.lower()

    if hasattr(logging, level_name):
        raise AttributeError("{} already defined in logging module".format(level_name))
    if hasattr(logging, method_name):
        raise AttributeError("{} already defined in logging module".format(method_name))
    if hasattr(logging.getLoggerClass(), method_name):
        raise AttributeError("{} already defined in logger class".format(method_name))

    def log_for_level(self: logging.Logger, message: str, *args: Any, **kwargs: Any) -> None:
        if self.isEnabledFor(level_num):
            self._log(level_num, message, args, **kwargs)

    def log_to_root(message: str, *args: Any, **kwargs: Any) -> None:
        logging.log(level_num, message, *args, **kwargs)

    logging.addLevelName(level_num, level_name)
    setattr(logging, level_name, level_num)
    setattr(logging.getLoggerClass(), method_name, log_for_level)
    setattr(logging, method_name, log_to_root)


class ResotoLogger(Logger):
    def debug2(self, msg: str, *args: Any, **kwargs: Any) -> None: ...  # noqa: E704

    def debug3(self, msg: str, *args: Any, **kwargs: Any) -> None: ...  # noqa: E704

    def debug4(self, msg: str, *args: Any, **kwargs: Any) -> None: ...  # noqa: E704

    def debug5(self, msg: str, *args: Any, **kwargs: Any) -> None: ...  # noqa: E704

    def trace(self, msg: str, *args: Any, **kwargs: Any) -> None: ...  # noqa: E704


def get_resoto_logger(name: Optional[str] = None) -> ResotoLogger:
    return getLogger(name)  # type: ignore


add_logging_level("DEBUG2", DEBUG2)
add_logging_level("DEBUG3", DEBUG3)
add_logging_level("DEBUG4", DEBUG4)
add_logging_level("DEBUG5", DEBUG5)
add_logging_level("TRACE", TRACE)

setLoggerClass(ResotoLogger)
setup_logger("resoto", force=False)
log = get_resoto_logger("resoto")



import argparse
import os
import shutil
import subprocess
from collections import defaultdict
from typing import List, Dict, Any, Union, Callable, Optional, Sequence, Tuple

DEFAULT_ENV_ARGS_PREFIX = "RESOTO_"


class Namespace(argparse.Namespace):
    def __getattr__(self, item: str) -> Any:
        return None


class _MachineHelpAction(argparse.Action):
    def __init__(
        self,
        option_strings: Sequence[str],
        dest: str = argparse.SUPPRESS,
        default: str = argparse.SUPPRESS,
        help: Optional[str] = None,
    ) -> None:
        super(_MachineHelpAction, self).__init__(
            option_strings=option_strings,
            dest=dest,
            default=default,
            nargs=0,
            help=help,
        )

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: Union[str, Sequence[Any], None],
        option_string: Optional[str] = None,
    ) -> None:
        parser.print_machine_help()  # type: ignore
        parser.exit()


class ArgumentParser(argparse.ArgumentParser):
    # Class variable containing the last return value of parse_args()
    # If parse_args() hasn't been called yet will return None for any
    # attribute.
    args = Namespace()

    def __init__(
        self,
        *args: Any,
        env_args_prefix: str = DEFAULT_ENV_ARGS_PREFIX,
        add_machine_help: bool = True,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.env_args_prefix = env_args_prefix
        self.add_machine_help = add_machine_help
        self.register("action", "machine_help", _MachineHelpAction)

        if self.add_machine_help:
            self.add_argument(
                "--machine-help",
                action="machine_help",
                help="print machine readable help",
            )

    def print_machine_help(self) -> None:
        for action in self._actions:
            if action.default == argparse.SUPPRESS:
                continue
            for option_string in action.option_strings:
                print(option_string)

    def parse_known_args(  # type: ignore
        self, args: Optional[Sequence[str]] = None, namespace: Optional[Namespace] = None
    ) -> Tuple[Namespace, List[str]]:
        for action in self._actions:
            env_name = None
            for option_string in action.option_strings:
                if option_string.startswith("--"):
                    env_name = self.env_args_prefix + option_string[2:].replace("-", "_").upper()
                    break
            if env_name is not None and action.default != argparse.SUPPRESS:
                new_default: Any = None
                if action.nargs not in (0, None):
                    new_default = os.environ.get(env_name)
                    if new_default is not None:
                        new_default = new_default.split(" ")
                    else:
                        new_default = []
                        for i in range(255):
                            new_ittr_default = os.environ.get(env_name + str(i))
                            if new_ittr_default is not None:
                                new_default.append(new_ittr_default)
                        if len(new_default) == 0:
                            new_default = None
                else:
                    new_default = os.environ.get(env_name)

                if new_default is not None:
                    if isinstance(action.type, type):
                        type_goal = action.type
                    elif callable(action.type):
                        type_goal = action.type
                    else:
                        type_goal = type(action.default)

                    if isinstance(new_default, list):
                        new_default = [convert(n, type_goal) for n in new_default]
                    else:
                        new_default = convert(new_default, type_goal)

                    action.default = new_default
        ret_args, ret_argv = super().parse_known_args(args=args, namespace=namespace)
        ArgumentParser.args = ret_args  # type: ignore
        return ret_args, ret_argv  # type: ignore


def get_arg_parser(
    add_help: bool = True,
    description: str = "resoto",
    env_args_prefix: str = DEFAULT_ENV_ARGS_PREFIX,
) -> ArgumentParser:
    arg_parser = ArgumentParser(description=description, add_help=add_help, env_args_prefix=env_args_prefix)
    return arg_parser


# removed from types in 3.0-3.9: introduced again in 3.10
NoneType = type(None)


def convert(value: Any, type_goal: Union[type, Callable[[Any], Any]]) -> Any:
    if type_goal is NoneType:  # type: ignore
        return value
    elif isinstance(type_goal, type):
        try:
            if type_goal in (str, int, float, complex):
                return type_goal(value)
            elif type_goal is bool:
                return value.lower() in ("true", "1", "yes")
            else:
                # don't know how to handle this type
                return value
        except Exception:
            # can not convert value
            return value
    elif callable(type_goal):
        return type_goal(value)


def args_dispatcher(dispatch_to: List[str], use_which: bool, argv: List[str]) -> Dict[str, List[str]]:
    dispatch_args = defaultdict(list)
    prog_args = defaultdict(list)

    for prog in dispatch_to:
        cmd = prog
        if use_which:
            cmd = shutil.which(prog)  # type: ignore
        result = subprocess.run([cmd, "--machine-help"], capture_output=True)
        if result.returncode != 0:
            continue
        for arg in result.stdout.decode("utf8").splitlines():
            prog_args[arg].append(prog)

    args = defaultdict(list)
    for i, arg in enumerate(argv):
        if arg.startswith("-"):
            arg_index = i + 1
            if len(argv) > arg_index:
                while not argv[arg_index].startswith("-"):
                    argval = str(argv[arg_index]).replace("'", "\\'")
                    args[arg].append(f"'{argval}'")
                    arg_index += 1
                    if len(argv) <= arg_index:
                        break
            if arg not in args:
                args[arg] = []

    for main_arg, arg_list in args.items():
        if main_arg in prog_args:
            for prog in prog_args[main_arg]:
                dispatch_args[prog].append(main_arg)
                dispatch_args[prog].extend(arg_list)

    return dict(dispatch_args)



import threading
import time
from contextlib import suppress, AbstractContextManager
from logging import Logger
from queue import Queue

from websocket import WebSocketApp, WebSocket  # type: ignore
import requests
import json
from concurrent.futures import ThreadPoolExecutor

from attr import define, evolve, field
from requests import Response

from resotolib.core.progress import Progress, ProgressDone
from resotolib.logger import log
from resotolib.event import EventType, remove_event_listener, add_event_listener, Event
from resotolib.args import ArgumentParser
from resotolib.jwt import encode_jwt_to_headers
from resotolib.core.ca import TLSData
from typing import Callable, Dict, Optional, List, Any

from resotolib.types import Json
from resotolib.utils import utc_str


@define(frozen=True)
class CoreFeedback:
    task_id: str
    step_name: str
    message_type: str
    core_messages: Queue[Json]
    context: List[str] = field(factory=list)

    def progress_done(self, name: str, current: int, total: int, context: Optional[List[str]] = None) -> None:
        self.progress(ProgressDone(name, current, total, path=context or self.context))

    def progress(self, progress: Progress) -> None:
        message = {
            "kind": "action_progress",
            "message_type": self.message_type,
            "data": {
                "task": self.task_id,
                "step": self.step_name,
                "context": self.context,
                "progress": progress.to_json(),
                "at": utc_str(),
            },
        }
        self.core_messages.put(message)

    def info(self, message: str, logger: Optional[Logger] = None) -> None:
        if logger:
            logger.warning(self.context_str + message)
        self._info_message("info", message)

    def error(self, message: str, logger: Optional[Logger] = None) -> None:
        if logger:
            logger.warning(self.context_str + message, exc_info=True)
        self._info_message("error", message)

    @property
    def context_str(self) -> str:
        return "[" + (":".join(self.context)) + "] " if self.context else ""

    def _info_message(self, level: str, message: str) -> None:
        self.core_messages.put(
            {
                "kind": "action_info",
                "message_type": self.message_type,
                "data": {
                    "task": self.task_id,
                    "step": self.step_name,
                    "level": level,
                    "message": self.context_str + message[0:500],  # truncate message to 500 characters
                },
            }
        )

    def with_context(self, *context: str) -> "CoreFeedback":
        return evolve(self, context=list(context))

    def child_context(self, *context: str) -> "CoreFeedback":
        return self.with_context(*(self.context + list(context)))


class SuppressWithFeedback(AbstractContextManager[None]):
    def __init__(self, message: str, feedback: CoreFeedback, logger: Optional[Logger] = None) -> None:
        self.message = message
        self.feedback = feedback
        self.logger = logger

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> Optional[bool]:
        if exc_type is not None:
            self.feedback.error(f"{self.message}: {exc_val}", self.logger)
            return True  # suppress exception
        return None


class CoreActions(threading.Thread):
    def __init__(
        self,
        identifier: str,
        resotocore_uri: str,
        resotocore_ws_uri: str,
        actions: Dict[str, Json],
        incoming_messages: Optional[Queue[Json]] = None,
        message_processor: Optional[Callable[[Json], Any]] = None,
        tls_data: Optional[TLSData] = None,
        max_concurrent_actions: int = 5,
    ) -> None:
        super().__init__()
        self.identifier = identifier
        self.resotocore_uri = resotocore_uri
        self.resotocore_ws_uri = resotocore_ws_uri
        self.actions = actions
        self.message_processor = message_processor
        self.ws: Optional[WebSocketApp] = None
        self.incoming_messages = incoming_messages
        self.tls_data = tls_data
        self.shutdown_event = threading.Event()
        # one thread is taken by the queue listener
        self.executor = ThreadPoolExecutor(max_workers=max_concurrent_actions + 1, thread_name_prefix=self.identifier)
        self.__connected = False

    def connected(self) -> bool:
        return self.__connected

    def run(self) -> None:
        def listen_on_queue(in_messages: Queue[Json]) -> None:
            while not self.shutdown_event.is_set():
                with suppress(Exception):
                    message = in_messages.get(timeout=1)
                    log.debug("Got feedback message. Send it to core", message)
                    if self.ws:
                        self.ws.send(json.dumps(message))

        self.name = self.identifier
        add_event_listener(EventType.SHUTDOWN, self.shutdown)
        if self.incoming_messages:
            self.executor.submit(listen_on_queue, self.incoming_messages)
        while not self.shutdown_event.is_set():
            log.debug("Connecting to resotocore message bus")
            try:
                self.connect()
            except Exception as e:
                log.error(e)
            time.sleep(1)

    def wait_for_ws(self, timeout: int = 10) -> bool:
        start = time.time()
        while self.ws is None and time.time() - start < timeout and not self.shutdown_event.is_set():
            time.sleep(0.1)
        return self.ws is not None

    def connect(self) -> None:
        for event, data in self.actions.items():
            if not isinstance(data, dict):
                data = None
            self.register(event, data)

        ws_uri = f"{self.resotocore_ws_uri}/subscriber/{self.identifier}/handle"
        log.debug(f"{self.identifier} connecting to {ws_uri}")
        headers: Dict[str, str] = {}
        if getattr(ArgumentParser.args, "psk", None):
            encode_jwt_to_headers(headers, {}, ArgumentParser.args.psk)
        try:
            self.ws = WebSocketApp(
                ws_uri,
                header=headers,
                on_open=self.on_open,
                on_message=self.on_message,
                on_error=self.on_error,
                on_close=self.on_close,
                on_ping=self.on_ping,
                on_pong=self.on_pong,
            )
            sslopt: Dict[Any, Any] = {}
            if self.tls_data:
                sslopt = {"ca_certs": self.tls_data.ca_cert_path}
            self.ws.run_forever(sslopt=sslopt, ping_interval=20, ping_timeout=10, ping_payload="ping")
        finally:
            self.ws = None

    def shutdown(self, _: Optional[Event] = None) -> None:
        remove_event_listener(EventType.SHUTDOWN, self.shutdown)
        log.debug("Received shutdown event - shutting down resotocore message bus listener")
        self.shutdown_event.set()
        self.executor.shutdown(wait=False, cancel_futures=True)
        if self.ws:
            self.ws.close()

    def register(self, action: str, data: Optional[Dict[str, str]] = None) -> bool:
        log.debug(f"{self.identifier} registering for {action} actions ({data})")
        return self.registration(action, requests.post, data)

    def unregister(self, action: str, data: Optional[Dict[str, str]] = None) -> bool:
        log.debug(f"{self.identifier} unregistering from {action} actions ({data})")
        return self.registration(action, requests.delete, data)

    def registration(
        self,
        action: str,
        client: Callable[..., Response],
        data: Optional[Dict[str, str]] = None,
    ) -> bool:
        url = f"{self.resotocore_uri}/subscriber/{self.identifier}/{action}"
        headers = {"accept": "application/json"}

        if getattr(ArgumentParser.args, "psk", None):
            encode_jwt_to_headers(headers, {}, ArgumentParser.args.psk)

        verify = None
        if self.tls_data:
            verify = self.tls_data.ca_cert_path

        r = client(url, headers=headers, params=data, verify=verify)
        if r.status_code != 200:
            raise RuntimeError(f'Error during (un)registration for "{action}"' f" actions: {r.content.decode('utf-8')}")
        return True

    def on_message(self, _: WebSocket, message: str) -> None:
        self.executor.submit(self.process_message, message)

    def process_message(self, message: str) -> None:
        try:
            json_message: Json = json.loads(message)
        except json.JSONDecodeError:
            log.exception(f"Unable to decode received message {message}")
            return
        log.debug(f"{self.identifier} received: {message}")
        if self.message_processor is not None and callable(self.message_processor):
            try:
                result: Json = self.message_processor(json_message)
                if result is None:
                    return
                if self.wait_for_ws() and self.ws:
                    log.debug(f"Sending reply {result}")
                    self.ws.send(json.dumps(result))
                else:
                    log.error(f"Unable to send reply {result}")
            except Exception:
                log.exception(f"Something went wrong while processing {message}")

    def on_error(self, _: WebSocket, e: Exception) -> None:
        log.debug(f"{self.identifier} message bus error: {e!r}")

    def on_close(self, _: WebSocket, close_status_code: int, close_msg: str) -> None:
        self.__connected = False
        log.debug(f"{self.identifier} disconnected from resotocore message bus: {close_status_code}: {close_msg}")

    def on_open(self, _: WebSocket) -> None:
        self.__connected = True
        log.debug(f"{self.identifier} connected to resotocore message bus")

    def on_ping(self, _: WebSocket, message: str) -> None:
        log.debug(f"{self.identifier} actions ping from resotocore message bus")

    def on_pong(self, _: WebSocket, message: str) -> None:
        log.debug(f"{self.identifier} actions pong from resotocore message bus")

    @staticmethod
    def add_args(arg_parser: ArgumentParser) -> None:
        arg_parser.add_argument(
            "--resotocore-subscriber-id",
            help="resotocore actions subscriber identifier (default: worker)",
            default="worker",
            dest="resotocore_subscriber_id",
        )



import asyncio
from asyncio import Queue, Future
from contextlib import suppress
from typing import (
    Callable,
    Awaitable,
    AsyncContextManager,
    Any,
    Optional,
    Sequence,
    Dict,
    Tuple,
    TypeVar,
)
from uuid import uuid1

import jsons
from aiohttp import WSMessage, WSMsgType
from aiohttp.web import Request, WebSocketResponse

from resotolib.logger import log

WSHandler = Dict[str, Tuple[Future[Any], WebSocketResponse]]
T = TypeVar("T")


async def clean_ws_handler(ws_id: str, websocket_handler: WSHandler) -> None:
    with suppress(Exception):
        handler = websocket_handler.get(ws_id)
        if handler:
            websocket_handler.pop(ws_id, None)
            future, ws = handler
            future.cancel()
            log.info(f"Cleanup ws handler: {ws_id} ({len(websocket_handler)} active)")
            if not ws.closed:
                await ws.close()


def js_str(a: Any) -> str:
    return jsons.dumps(a, strip_privates=True)  # type: ignore


async def accept_websocket(
    request: Request,
    *,
    handle_incoming: Callable[[str], Awaitable[None]],
    websocket_handler: WSHandler,
    outgoing_context: Optional[Callable[[], AsyncContextManager[Queue[T]]]] = None,
    initial_messages: Optional[Sequence[Any]] = None,
    outgoing_fn: Callable[[T], str] = js_str,
) -> WebSocketResponse:
    ws = WebSocketResponse(autoping=True, heartbeat=20)
    await ws.prepare(request)
    wsid = str(uuid1())

    # in case we wait for an initial authorization message, only wait for a limited amount of tine
    async def wait_for_authorization() -> None:
        counter = 10
        while request.get("authorized", False) is not True and counter >= 0:
            await asyncio.sleep(1)
            counter -= 1
        if counter <= 0:
            log.info(f"Wait for authorization: message listener {wsid}: Timeout. Hang up.")
            await clean_ws_handler(wsid, websocket_handler)

    async def receive() -> None:
        try:
            async for msg in ws:
                if isinstance(msg, WSMessage) and msg.type in (
                    WSMsgType.ERROR,
                    WSMsgType.CLOSE,
                    WSMsgType.CLOSED,
                ):
                    break
                elif isinstance(msg, WSMessage) and msg.type == WSMsgType.TEXT and len(msg.data.strip()) > 0:
                    log.debug(f"Incoming message: type={msg.type} data={msg.data} extra={msg.extra}")
                    await handle_incoming(msg.data)
        except Exception as ex:
            # do not allow any exception - it will destroy the async fiber and cleanup
            log.info(f"Receive: message listener {wsid}: {ex}. Hang up.")
        finally:
            await clean_ws_handler(wsid, websocket_handler)

    async def send(ctx: Callable[[], AsyncContextManager[Queue[T]]]) -> None:
        try:
            # wait for the request to become authorized, before we will send any message
            while request.get("authorized", False) is not True:
                await asyncio.sleep(1)
            # send all initial messages
            if initial_messages:
                for msg in initial_messages:
                    await ws.send_str(outgoing_fn(msg) + "\n")
            # attach to the queue and wait for messages
            async with ctx() as events:
                while True:
                    event = await events.get()
                    await ws.send_str(outgoing_fn(event) + "\n")
        except Exception as ex:
            # do not allow any exception - it will destroy the async fiber and cleanup
            log.info(f"Send: message listener {wsid}: {ex}. Hang up.")
        finally:
            await clean_ws_handler(wsid, websocket_handler)

    tasks = [asyncio.create_task(receive())]
    if outgoing_context is not None:
        tasks.append(asyncio.create_task(send(outgoing_context)))
    if request.get("authorized", False) is not True:
        tasks.append(asyncio.create_task(wait_for_authorization()))

    to_wait = asyncio.gather(*tasks)
    websocket_handler[wsid] = (to_wait, ws)
    await to_wait
    return ws
